#ifdef MT_LAPACK

#include "array.h"
#include "std.h"

#ifdef __cplusplus
extern "C"{
#endif

#include "cblas.h"
#ifdef MT_MSVC
#  include "blaswrap.h"
#endif
#include <f2c.h>
#undef small
#undef large
#include "clapack.h"
#undef max
#undef min
#undef abs

#ifdef __cplusplus
}
#endif

#if MT_MSVC
void blas_MM(arr& X,const arr& A,const arr& B){       MT::useLapack=false; innerProduct(X,A,B); MT::useLapack=true; };
void blas_MsymMsym(arr& X,const arr& A,const arr& B){ MT::useLapack=false; innerProduct(X,A,B); MT::useLapack=true; };
void blas_Mv(arr& y,const arr& A,const arr& x){       MT::useLapack=false; innerProduct(y,A,x); MT::useLapack=true; };
#else
void blas_MM(arr& X,const arr& A,const arr& B){
  CHECK(A.d1==B.d0,"matrix multiplication: wrong dimensions");
  X.resize(A.d0,B.d1);
  cblas_dgemm(CblasRowMajor,
	      CblasNoTrans,CblasNoTrans,
	      A.d0,B.d1,A.d1,
	      1.,A.p,A.d1,
	      B.p,B.d1,
	      0.,X.p,X.d1);
#if 0//test
  MT::useLapack=false;
  std::cout  <<"blas_MM error = " <<maxDiff(A*B,X,0) <<std::endl;
  MT::useLapack=true;
#endif
}

void blas_Mv(arr& y,const arr& A,const arr& x){
  CHECK(A.d1==x.N,"matrix multiplication: wrong dimensions");
  y.resize(A.d0);
  cblas_dgemv(CblasRowMajor,
	      CblasNoTrans,
	      A.d0,A.d1,
	      1.,A.p,A.d1,
	      x.p,1,
	      0.,y.p,1);
#if 0 //test
  MT::useLapack=false;
  std::cout  <<"blas_Mv error = " <<maxDiff(A*x,y,0) <<std::endl;
  MT::useLapack=true;
#endif
}

void blas_MsymMsym(arr& X,const arr& A,const arr& B){
  CHECK(A.d1==B.d0,"matrix multiplication: wrong dimensions");
  X.resize(A.d0,B.d1);
  cblas_dsymm(CblasRowMajor,
	      CblasLeft,CblasUpper,
	      A.d0,B.d1,
	      1.,A.p,A.d1,
	      B.p,B.d1,
	      0.,X.p,X.d1);

#if 0 //test
  arr Y(A.d0,B.d1);
  uint i,j,k;
  Y.setZero();
  for(i=0;i<Y.d0;i++) for(j=0;j<Y.d1;j++) for(k=0;k<A.d1;k++)
    Y(i,j) += A(i,k) * B(k,j);
  std::cout  <<"blas_MsymMsym error = " <<sqrDistance(X,Y) <<std::endl;
#endif
}
#endif

void lapack_Ainv_b_sym(arr& x,const arr& A, const arr& b){
  static arr Acol;
  integer n=A.d0,m=1;
  integer info;
  x=b;
  Acol=A;
  dposv_((char*)"L", &n, &m, Acol.p, &n, x.p, &n, &info);
  CHECK(!info,"lapack_Ainv_b_sym error info = " <<info);
  
#if 0
  arr y = inverse(A)*b;
  std::cout  <<"lapack_invA_b_sym error = " <<sqrDistance(x,y) <<std::endl;
#endif
}

uint lapack_SVD(const arr& A,
	       arr& U,
	       arr& d,
	       arr& Vt){
  static arr Atmp,work;
  Atmp=A;
  //transpose(Atmp,A);
  integer M=A.d0,N=A.d1,D=M<N?M:N;
  U.resize(M,D);
  d.resize(D);
  Vt.resize(D,N);
  work.resize(10*(M+N));
  integer info,wn=work.N;
  dgesvd_((char*)"S", (char*)"S", &N, &M, Atmp.p, &N, d.p, Vt.p, &N, U.p, &D, work.p,&wn, &info);
  CHECK(!info,"LAPACK SVD error info = " <<info);
  return D;
}

void lapack_LU(const arr& A, arr& LU){
  LU = A;
  integer M=A.d0,N=A.d1,D=M<N?M:N,info;
  intA piv(D);
  dgetrf_(&N, &M, LU.p, &N, (integer*)piv.p, &info);
  CHECK(!info,"LAPACK SVD error info = " <<info);
}

void lapack_EigenDecomp(const arr& symmA, arr& Evals, arr& Evecs) {
	CHECK(symmA.nd==2 && symmA.d0==symmA.d1,"not symmetric");
 	static arr work;
  Evecs=symmA;
  integer N=symmA.d0;
  Evals.resize(N);
  Evecs.resize(N,N);
  // any number for size
  work.resize(10*(3*N));
  integer info, wn=work.N;
  dsyev_((char*)"V", (char*)"U", &N, Evecs.p, 
  	&N, Evals.p, work.p, &wn, &info);
  transpose(Evecs);
  CHECK(!info,"lapack_EigenDecomp error info = " <<info);
}

bool lapack_isPositiveSemiDefinite(const arr& symmA) {
  // Check that all eigenvalues are nonnegative.
  arr d, V;
  lapack_EigenDecomp(symmA, d, V);
  // d is nondecreasing ??!??
  uint i;
  FOR1D(d, i)
    if (d(i) < 0.)
      return false;
    
    return true;
}

void lapack_inverseSymPosDef(arr& invA,const arr& A){
  CHECK(A.d0==A.d1,"");
  integer n=A.d0;
  integer info;
  invA=A;
  //compute cholesky
  dpotrf_((char*)"L", &n, invA.p, &n, &info);
  CHECK(!info,"LAPACK Cholesky decomp error info = " <<info);
  //invert
  dpotri_((char*)"L", &n, invA.p, &n, &info);
  CHECK(!info,"lapack_inverseSymPosDef error info = " <<info);
  uint i,j;
  for(i=0;i<(uint)n;i++) for(j=0;j<i;j++) invA(i,j)=invA(j,i);
}

/*
dpotri uses:
dtrtri = invert triangular

dlauum = multiply L'*L
*/

#else
#ifndef MT_MSVC
#  warning "MT_LAPACK undefined - using inefficient implementations"
#endif
void blas_MM(arr& X,const arr& A,const arr& B){ innerProduct(X,A,B); };
void blas_MsymMsym(arr& X,const arr& A,const arr& B){ innerProduct(X,A,B); };
uint lapack_SVD(const arr& A, arr& U, arr& d, arr& Vt){NIY};
void lapack_LU(const arr& A, arr& LU){NIY};
void lapack_EigenDecomp(const arr& symmA, arr& Evals, arr& Evecs){NIY};
bool lapack_isPositiveSemiDefinite(const arr& symmA){NIY};
void lapack_inverseSymPosDef(arr& invA,const arr& A){NIY};
#endif
