/*  Copyright (C) 2000, 2006  Marc Toussaint (mtoussai@inf.ed.ac.uk)
under the terms of the GNU LGPL (http://www.gnu.org/copyleft/lesser.html)
see the `std.h' file for a full copyright statement  */

#include "infer.h"
#include "infer_tree.h"
#include <algorithm>

#define LOG_NORM_SCALE
int DEBUG_INFER_LEVEL=0;
#define DEBUG_INFER(l,x) if(l<=DEBUG_INFER_LEVEL){ x; }


//===========================================================================
//
// global objects
//

struct iSpace{
  VarA variables;
  FacA factors;
  
  iSpace();
  void deleteAll();
  void write(std::ostream& os) const;
  void writeVariables(std::ostream& os) const;
};

iSpace globalInferSpace;
void checkSpaceRegistry(iSpace *S=0);


//===========================================================================
//
// Declarations of various minor functions
//


void get_vars(VarA& V, const FacA& factors);
void getNeighbors(Factor* f, FacA& neighbors);
void convert_id2config(uintA& config, uint id, const uintA& varDimensions);
uint convert_config2id(uintA& config, const uintA& varDimensions);
uint convert_config2id(boolA truths);
int get_list_index(FactorGraph& fg, Factor* f);
uint get_list_index_unsigned(FactorGraph& fg, Factor* f);

void get_vars(VarA& V, const FacA& factors){
  V.clear();
  uint i, k;
  FOR1D(factors, i){
    FOR1D(factors(i)->varIds, k){
      V.setAppend(getvar(factors(i)->varIds(k)));
    }
  }
}

void getNeighbors(Factor* f, FacA& neighbors){
  neighbors.clear();
  uint i;
  FOR1D(f->msg_pairs, i){
    if(f->msg_pairs(i)->f1 == f)
      neighbors.append(f->msg_pairs(i)->f2);
    else
      neighbors.append(f->msg_pairs(i)->f1);
  }
}

MessagePair* getMessagePair(const Factor* f1, const Factor* f2){
  uint i;
  FOR1D(f1->msg_pairs, i){
    if(f1->msg_pairs(i)->f1 == f2  ||  f1->msg_pairs(i)->f2 == f2)
      return f1->msg_pairs(i);
  }
  return NULL;
}

Factor* getMessage(const Factor* f_from, const Factor* f_to){
  MessagePair* msg_pair = getMessagePair(f_from, f_to);
  CHECK(msg_pair != NULL,  "Message pair not found");
  if(msg_pair->f1 == f_from)
    return &msg_pair->m12;
  else
    return &msg_pair->m21;
}

void writeMessage(const Factor* f_from, const Factor* f_to, ostream& os){
  os <<*getMessage(f_from, f_to);
}


int get_list_index__orig(FactorGraph& fg, Factor* f_orig){
  return fg.F.findValue(f_orig);
}

uint get_list_index_unsigned__orig(FactorGraph& fg, Factor* f_orig){
  uint i=0;
  while(i < fg.F.N){
    if(fg.F(i) == f_orig)
      break;
    i++;
  }
  CHECK(i<fg.F.N, "Factor over " <<f_orig->varIds <<" not found in FactorGraph!" <<endl);
  return i;
}





//===========================================================================
//
// space registry stuff
//

//[[ als deinit members von Variable und Factor ]]
void unregisterVariable(Variable& v){
  //CHECK(v.factors.N==0,"can't unregister variable which is still linked to factors");
  if(v.factors.N) MT_MSG("shouldn't unregister variable which is still linked to factors");
  if(!v.space) return;
  v.space->variables.removeValue(&v);
}

void unregisterFactor(Factor& f){
  if(!f.space) return;
  for(uint i=0;i<f.varIds.N;i++){
    //f.space->variables(f.varIds(i))->factors.removeValue(&f); [[REDO!]]
  }
  f.space->factors.removeValue(&f);
  f.space=0;
}

void checkSpaceRegistry(iSpace *S){
  if(!S) S=&globalInferSpace;
  Variable *v;
  Factor *f;
  uint i,j,k;
  for(i=0;i<S->variables.N;i++){
    v=S->variables(i);
    CHECK(v->space==S,"variale points to another space");
    CHECK(v->id==i,"variable id is not the position space->variables registry");
    for(j=0;j<v->factors.N;j++){
      f=v->factors(j);
      k=f->varIds.findValue(i);
      CHECK(k<f->varIds.N,"variable's factor doesn't include the variable's id");
      CHECK(v->dim==f->dim(k),"variable's factor thinks variable has different dim");
    }
  }
  for(i=0;i<S->factors.N;i++){
    f=S->factors(i);
    CHECK(f->space==S,"factor points to another space");
    for(j=0;j<f->varIds.N;j++){
      v=S->variables(f->varIds(j));
      k=v->factors.findValue(f);
      //CHECK(k<v->factors.N,"factor's variable doesn't include the factor in its list"); [[REDO!]]
      CHECK(f->dim(j)==v->dim,"factor's variable thinks it is of different dim");
    }
  }
}



//===========================================================================
//
// iSpace
//


iSpace::iSpace(){
  variables.memMove=true;
  factors.memMove=true;
}

void iSpace::deleteAll(){
  Factor *f;
  while(factors.N){
    f=factors(0);
    while(f->msg_pairs.N) delete f->msg_pairs(0);
    delete f;
  }
  while(variables.N) delete variables(0);
}

void iSpace::write(std::ostream& os) const{
  os <<"<iSpace>"
    <<" #variable="   <<variables.N <<endl;
  //<<" #msg_pairs=" <<g.msg_pairs.N <<endl;
  uint i,j;
  for(i=0;i<variables.N;i++){
    os <<'[' <<i <<"] " <<variables(i)->name <<' ' <<variables(i)->dim;
    for(j=0;j<variables(i)->factors.N;j++) os <<',' <<variables(i)->factors(j)->varIds;
    os <<endl;
    CHECK(variables(i)->id==i,"");
  }
  os <<" #factors="    <<factors.N <<endl;
  for(i=0;i<factors.N;i++){
    os <<'[' <<i <<"] " <<*factors(i);
    //for(j=0;j<factors(i)->varIds.N;j++) os <<',' <<variables(factors(i)->varIds(j))->name;
    os <<endl;
  }
  os <<endl;
  /*
  for(i=0;i<g.msg_pairs.N;i++){
  os <<'[' <<g.msg_pairs(i).from <<'-' <<g.msg_pairs(i).to <<"] " <<g.msg_pairs.elem(i) <<endl;
  }
  */
  os <<"\n</iSpace>" <<endl;
}


void iSpace::writeVariables(std::ostream& os) const {
  uint i;
  for(i=0; i<variables.N; i++){
    os <<"[" <<variables(i)->id <<"]  " <<variables(i)->name <<endl;
  }
}





//===========================================================================
//
// Variable
//

Variable::Variable(){
  space=0;
  factors.memMove=true;
}

Variable::Variable(uint _dim,const char *_name){
  space=0;
  init(0,_dim,_name);
}

Variable::~Variable(){
  unregisterVariable(*this);
}

void Variable::init(iSpace *_space,uint _dim,const char *_name){
  factors.memMove=true;
  if(!_space) _space=&globalInferSpace;
  unregisterVariable(*this);
  space=_space;
  space->variables.append(this);
  id=space->variables.N-1;
  dim=_dim;
  name=_name;
}

void Variable::write(ostream& os) const{
  os <<name <<"[" <<id <<"]" <<dim;
}







//===========================================================================
//
// Factor
//

Factor::Factor(){
  space=NULL;
  msg_pairs.memMove=true;
  logP=0.;
}

Factor::~Factor(){
  //     cout<<this<<endl;
  //     cout<<"Killing factor over ";printvars(varIds);cout<<endl;
  //     cout<<P<<endl;
  //     cout<<"logP="<<(logP)<<endl;
  unregisterFactor(*this);
}

Factor::Factor(const uintA& vars){
  space=NULL;
  msg_pairs.memMove=true;
  init(0,vars);
  setOne();
}

Factor::Factor(const uintA& vars,const char* text){
  space=0;
  msg_pairs.memMove=true;
  init(0,vars);
  setText(text);
}

Factor::Factor(const uintA& vars,const arr& q){
  space=0;
  msg_pairs.memMove=true;
  init(0,vars);
  set(q);
}

void Factor::init(iSpace *_space,const uintA& variables){
  if(!_space) _space=&globalInferSpace;
  unregisterFactor(*this);
  space=_space;
  space->factors.append(this);
  varIds=variables;
  dim.resize(varIds.N);
  for(uint i=0;i<varIds.N;i++){
    dim(i)=space->variables(varIds(i))->dim;
    //space->variables(varIds(i))->factors.append(this); [[REDO!]]
  }
}

void Factor::operator=(const Factor& q){
  init(q.space,q.varIds);
  P=q.P; logP=q.logP;
}

void Factor::set(const arr& p){
  P = p;
  CHECK(P.N==product(dim),"Factor set with ill-dimensioned array");
  P.reshape(dim);
  logP = 0.;
#ifdef LOG_NORM_SCALE
  lognormScale(P,logP);
#endif
}

void Factor::setText(const char* str){
  P.setText(str);
  CHECK(P.N==product(dim),"Factor set with ill-dimensioned array");
  P.reshape(dim);
  logP = 0.;
#ifdef LOG_NORM_SCALE
  lognormScale(P,logP);
#endif
}

void Factor::setOne(){
  P.resize(dim);
  P.setUni(1.);
  logP=0.;
#ifdef LOG_NORM_SCALE
  lognormScale(P,logP);
#endif
}

void Factor::setUniform(){
  double normalizedValue = 1./product(dim);
  //uint i;
  //FOR1D(dim, i)
  //  normalizedValue /= dim(i);
  P.setUni(normalizedValue);
  logP=0.;
#ifdef LOG_NORM_SCALE
  lognormScale(P,logP);
  CHECK(logP == 0., "Normalizing in lognormScale failed.");
#endif
}

void Factor::setRandom(){
  rndUni(P,.1,1.,false);
  logP=rnd.uni();
#ifdef LOG_NORM_SCALE
  lognormScale(P,logP);
#endif
}

void Factor::setEvidence(uint e){
  P=0.;
  P(e)=1.;
  logP=0.;
}

bool Factor::operator==(Factor& q){
  //uint i=P.maxIndex();
  //double ratio=P(i)/q.P(i);
  lognormScale(P,logP,true);
  lognormScale(q.P,q.logP,true);
  bool c1 = varIds==q.varIds;
  bool c2 = dim==q.dim;
  bool c3 = maxDiff(P,q.P,0)<1e-10;
  bool c4 = fabs(logP-q.logP)/(1+fabs(logP)+fabs(q.logP))<1e-10;
  bool r = c1  &&  c2  &&  c3  &&  c4;
  //if(!r){
    //MT_MSG("unequal table factors: " <<id <<q.id <<dim <<q.dim <<maxDiff(P,q.P) <<' ' <<logP <<' ' <<q.logP);
  //}
  return r;
}

void Factor::getP(arr& p) const{
  p = ::exp(logP)*P;
}

void Factor::write(std::ostream& os) const{
  os <<"factor over " <<varIds <<"  ";
  uint v;
  FOR1D(varIds, v) os <<getvar(varIds(v))->name <<' ';
  os <<endl;
  if(dim.N>12 || P.N>100)
    os <<"prob table omitted since too large (logP=" <<logP <<")" <<endl;
  else{
    os <<" P";
    P.write(os," ","\n   ");
    os <<'*' <<::exp(logP) <<"=exp(" <<logP <<") " <<endl;
  }
}

void Factor::writeNice(std::ostream& os) const{
  uint i;
  os <<"[ ";
  FOR1D(varIds, i)
    os<<getvar(varIds(i))->name<<"  ";
  os <<"]";
}

void Factor::writeExtremelyNice(std::ostream& os) const {
  writeNice(os);os<<endl;
  uint i, k;
  for(i=0;i<P.N;i++){
    os<<i<<": "<<P.elem(i)<<" (*"<<exp(logP)<<")     ";
    uintA config;
    P.getIndexTuple(config, i);
    FOR1D(varIds, k){
      os<<getvar(varIds(k))->name<<"="<<config(k)<<"   ";
    }
    os<<endl;
  }
}

void Factor::checkCondNormalization(uint left,double tol){
  tensorCheckCondNormalization_with_logP(P,left, logP, tol);
}



//===========================================================================
//
// MessagePair
//

MessagePair::MessagePair(){
  f1=f2=NULL;
}

MessagePair::MessagePair(Factor *_f1,Factor *_f2){
  init(_f1,_f2);
}

MessagePair::~MessagePair(){
  f1->msg_pairs.removeValue(this);
  f2->msg_pairs.removeValue(this);
}

void MessagePair::init(Factor *_f1,Factor *_f2){
  f1=_f1;
  f2=_f2;
  setSection(varIds, f1->varIds, f2->varIds);
  CHECK(f1->space==f2->space,"");
  m12.init(f1->space,varIds);  m12.setOne();
  m21.init(f1->space,varIds);  m21.setOne();
  f1->msg_pairs.append(this);
  f2->msg_pairs.append(this);
}

void MessagePair::write(ostream& os) const{
  os <<varIds <<" [ " <<f1->varIds <<" --- " <<f2->varIds <<" ]  "/*<<f1<<"  "<<f2*/<<endl;
  //   os <<"m12=\n" <<m12 <<"m21=\n" <<m21 <<endl;
}

void MessagePair::writeIds(ostream& os) const{
  MT::IOraw=true;
  os <<"[" <<f1->varIds <<" |" <<f2->varIds <<" ]";
  MT::IOraw=false;
}


void checkMessagePairConsistency(MT::Array<MessagePair*> msg_pairs){
  uint i, k;
  int idx;
  Factor* f;
  FOR1D(msg_pairs, i){
    f = msg_pairs(i)->f1;
    FOR1D(f->msg_pairs, k){
      idx = msg_pairs.findValue(f->msg_pairs(k));
      CHECK(idx>=0, "Factor references an unknown message pair!");
      CHECK(f->msg_pairs(k)->f1 == f  ||  f->msg_pairs(k)->f2 == f, "Factor references a foreign message pair!");
    }
    f = msg_pairs(i)->f2;
    FOR1D(f->msg_pairs, k){
      idx = msg_pairs.findValue(f->msg_pairs(k));
      CHECK(idx>=0, "Factor references an unknown message pair!");
      CHECK(f->msg_pairs(k)->f1 == f  ||  f->msg_pairs(k)->f2 == f, "Factor references a foreign message pair!");
    }
  }
}


//===========================================================================
//
// FactorGraph
//


void FactorGraph::write(std::ostream& os, bool writeBeliefs, bool writeMessages) const{
  uint i;
  os <<"<FactorGraph>"<<endl;
  //<<" #variable="   <<g.variables.dims.N
  //        <<" #factors="    <<this->factors.N
  //             <<" B_v="    <<this->B_v.N
  //        <<" #msg_pairs=" <<this->msg_pairs.N <<endl;
  //os <<g.variables <<endl;
  os <<"Original factors (" <<this->F.N <<"):" <<endl;
  for(i=0;i<this->F.N;i++){
    os <<"[" <<i <<"] " <<this->F(i)->varIds<<"  "<<this->F(i)<<endl;
//     os <<*F(i);
    F(i)->writeExtremelyNice(os);
//     os <<"msg: "<<F(i)->msg_pairs <<endl;
  }
  os <<endl;
  if(writeBeliefs){
    os <<"Beliefs-cliques (" <<this->B_c.N <<"):" <<endl;
    for(i=0;i<this->B_c.N;i++){
      os <<"[" <<i <<"] " <<this->B_c(i)->varIds <<endl;
      os <<*B_c(i);
    }
    os <<endl;
    os <<"Beliefs-vars (" <<this->B_v.N <<"):" <<endl;
    for(i=0;i<this->B_v.N;i++){
      os <<"[" <<i <<"] " <<this->B_v(i)->varIds <<endl;
      os <<*B_v(i);
    }
    os <<endl;
  }
  if(writeMessages){
    os <<"msg_pairs (" <<this->msg_pairs.N <<"):" <<endl;
    //     uint i1, i2;
    for(i=0;i<this->msg_pairs.N;i++){
      os <<"[" <<i <<"] " <<this->msg_pairs(i)<<"  ";
      //       FOR1D(factors, i1){
      //           if(factors(i1)==msg_pairs(i)->f1) break;
      //       }
      //       FOR1D(factors, i2){
      //           if(factors(i2)==msg_pairs(i)->f2) break;
      //       }
      //       os<<i1<<"-"<<i2<<"  ";
      os<<this->msg_pairs(i)->varIds  <<" [ " <<this->msg_pairs(i)->f1->varIds <<"---" <<this->msg_pairs(i)->f2->varIds <<"]  " <<msg_pairs(i)->f1<<"  "<<msg_pairs(i)->f2 <<endl;
      os<<"m12  "<<msg_pairs(i)->m12;
      os<<"m21  "<<msg_pairs(i)->m21;
    }
  }
  os <<"</FactorGraph>" <<endl;
}


void FactorGraph::writeNice(std::ostream& os) const {
  uint i, j;
  os <<"Factors (specified by their vars):" <<endl;
  for(i=0;i<this->B_c.N;i++){
    os <<"[" <<i <<"] "<<endl;
    os<<this->B_c(i)->varIds.N<<"  ";
    os <<this->B_c(i)->varIds <<endl;
    FOR1D(B_c(i)->varIds, j){os <<getvar(B_c(i)->varIds(j))->name<<"  ";}
    os<<endl;
  }
}


void FactorGraph::writeMessagePairs(std::ostream& os) const {
  uint s;
  FOR1D(this->msg_pairs, s){
    os <<*this->msg_pairs(s);
  }
}


void FactorGraph::writeVariableBeliefs(std::ostream& os) const{
  uint i;
  os <<"Variable beliefs:"<<endl;
  FOR1D(B_v, i){
    os <<*B_v(i)<<endl;
  }
}


void FactorGraph::checkCondNormalization_B_c(double tol){
  uint i;
  FOR1D(B_c, i){
    B_c(i)->checkCondNormalization(1,tol);
  }
}


void FactorGraph::setCliqueBeliefs(const FacA& fs_orig){
  uint i;
  // build belief factors if needed
  Factor* b;
  if(B_c.N == 0  &&  fs_orig.N > 0){
    FOR1D(fs_orig, i){
      b = new Factor(fs_orig(i)->varIds);
      B_c.append(b);
    }
//     MT_MSG("Rebuilding clique beliefs");
  }
  CHECK(fs_orig.N == B_c.N, "Number of original factors does not fit number of clique beliefs.");
  FOR1D(B_c, i){
    B_c(i)->P = fs_orig(i)->P;
    B_c(i)->logP = fs_orig(i)->logP;
  }
}


void FactorGraph::resetMessages(){
  uint i;
  FOR1D(msg_pairs, i){
    msg_pairs(i)->m12.setOne();
    msg_pairs(i)->m21.setOne();
  }
}


void FactorGraph::resetVariableFactors(){
  uint i;
  FOR1D(F_v, i){
    F_v(i)->setOne();
  }
}

void FactorGraph::resetVariableBeliefs(){
  uint i;
  FOR1D(B_v, i){
    B_v(i)->setOne();
  }
}


void FactorGraph::deleteAll(){
  uint i;
  FOR1D(msg_pairs, i) delete msg_pairs(i);
  //FOR1D(F, i)   delete F(i);
  FOR1D(F_v, i) delete F_v(i);
  FOR1D(B_c, i) delete B_c(i);
  FOR1D(B_v, i) delete B_v(i);
}


double FactorGraph::computeBeliefs(){
  double change, maxChange = 0.0;
  arr P_old;
  uint i;
  FOR1D(F, i){
    P_old = B_c(i)->P;
    collectBelief(*B_c(i), *F(i), 0);
    change = (B_c(i)->P - P_old).absMax();
    if(change > maxChange) maxChange = change;
  }
  FOR1D(F_v, i){
    P_old = B_v(i)->P;
    collectBelief(*B_v(i), *F_v(i), 0);
    change = (B_v(i)->P - P_old).absMax();
    if(change > maxChange) maxChange = change;
  }
  return maxChange;
}


Factor* FactorGraph::getBelief(Factor* f_orig){
  int idx;
  idx = F.findValue(f_orig); // TODO teuer, da sehr haeufig aufgerufen!
  if(idx >= 0)
    return B_c(idx);
  else {
    idx = F_v.findValue(f_orig);
    CHECK(idx>=0, "unknown factor "<<f_orig);
    return B_v(idx);
  }
}

void FactorGraph::setF2Bmap(){
  if(!F2B.empty()) F2B.clear();
  uint i;
  FOR1D(F, i){F2B[F(i)] = B_c(i);}
  FOR1D(F_v, i){F2B[F_v(i)] = B_v(i);}
}

void FactorGraph::addF2Bmap(Factor* f, Factor* b){
  F2B[f] = b;
}

Factor* FactorGraph::getBelief_fast(Factor* f_orig){
  return F2B[f_orig];
}

void FactorGraph::setV2Fv(){
  if(!V2Fv.empty()) V2Fv.clear();
  uint i, k;
  FOR1D(V, i){
    FOR1D(F_v, k){
      if(F_v(k)->varIds(0)==V(i)->id){
        V2Fv[V(i)->id] = F_v(k);
        break;
      }
    }
  }
}

void FactorGraph::setV2F(){
  if(!V2F.empty()) V2F.clear();
  uint i, k;
  FOR1D(V, i){
    FacA dummy;
    V2F[V(i)->id] = dummy;
    FOR1D(F, k){
      if(F(k)->varIds(0)==V(i)->id){
        V2F[V(i)->id].append(F(k));
      }
    }
  }
}

void FactorGraph::addV2Fmap(Factor* f){
  V2F[f->varIds(0)].append(f);
}


// void FactorGraph::checkFaithfulness(){
//   uint DEBUG=0;
//   if(DEBUG>0){cout<<"checkFaithfulness [START]"<<endl;}
//   uint i, k;
//   // for clique factors
//   if(DEBUG>0){cout<<"Checking normal factors:"<<endl;}
//   FOR1D(B_c, i){
//     if(DEBUG>0){cout<<"Belief is:"<<endl<<*B_c(i);}
//     Factor product(B_c(i)->varIds);
//     product.setOne();
//     if(DEBUG>0){cout<<"Multiplying:"<<endl;}
//     if(DEBUG>0){cout<<"* orig factor:"<<endl<<*F(i);}
//     tensorMultiply(product, *F(i));
//     FOR1D(B_c(i)->msg_pairs, k){
//       if(B_c(i)->msg_pairs(k)->f1 == B_c(i)){
//         if(DEBUG>0){cout<<"* msg:"<<endl<<B_c(i)->msg_pairs(k)->m21;}
//         tensorMultiply(product, B_c(i)->msg_pairs(k)->m21);
//       }
//       else if(B_c(i)->msg_pairs(k)->f2 == B_c(i)){
//         if(DEBUG>0){cout<<"* msg:"<<endl<<B_c(i)->msg_pairs(k)->m12;}
//         tensorMultiply(product, B_c(i)->msg_pairs(k)->m12);
//       }
//       else HALT("Impossible!");
//     }
//     if(DEBUG>0){cout<<"Product is:"<<endl<<product;}
//     if(!(product == *B_c(i))){
//       cerr<<"Product:"<<endl<<product<<endl;
//       cerr<<"*B_c(i):"<<endl<<*B_c(i)<<endl;
//       CHECK(product == *B_c(i), "Faithfulness failed for factor over "<<B_c(i)->varIds);
//     }
//   }
//   if(DEBUG>0){cout<<"Checking variable factors:"<<endl;}
//   // for variable factors
//   FOR1D(B_v, i){
//     if(DEBUG>0){cout<<"Belief is:"<<endl<<*B_v(i);}
//     Factor product(B_v(i)->varIds);
//     product.setOne();
//     if(DEBUG>0){cout<<"Multiplying:"<<endl;}
//     FOR1D(B_v(i)->msg_pairs, k){
//       if(B_v(i)->msg_pairs(k)->f1 == B_v(i)){
//         if(DEBUG>0){cout<<"* msg:"<<endl<<B_v(i)->msg_pairs(k)->m21;}
//         tensorMultiply(product, B_v(i)->msg_pairs(k)->m21);
//       }
//       else if(B_v(i)->msg_pairs(k)->f2 == B_v(i)){
//         if(DEBUG>0){cout<<"* msg:"<<endl<<B_v(i)->msg_pairs(k)->m12;}
//         tensorMultiply(product, B_v(i)->msg_pairs(k)->m12);
//       }
//       else HALT("Impossible!");
//     }
//     if(DEBUG>0){cout<<"Product is:"<<endl<<product;}
//     if(!(product == *B_v(i))){
//       cerr<<"Product:"<<endl<<product<<endl;
//       FOR1D(product.P, k){cout<<product.P.elem(k)<<endl;}
//       cerr<<"*B_v(i):"<<endl<<*B_v(i)<<endl;
//       FOR1D(B_v(i)->P, k){cout<<B_v(i)->P.elem(k)<<endl;}
//       CHECK(product == *B_v(i), "Faithfulness failed for factor over "<<B_v(i)->varIds);
//     }
//   }
//   if(DEBUG>0){cout<<"checkFaithfulness [END]"<<endl;}
// }


void get_vars(FactorGraph& fg, uintA& varIds){
  varIds.clear();
  uint i;
  FOR1D(fg.B_c, i) varIds.setAppend(fg.B_c(i)->varIds);
}


// Edge directions:
// edges between clique factors (cf) and variable factors (vf)
// cf-->vf  iff  variable in vf is the one that is conditioned on in cf (the first one there)
// cf<--vf  otherwise
#if 0
void LoopyBP::constructBipartiteFactorGraph(FactorGraph& fg, const FacA& factors){
  uint i,j;
  //copy factors
  fg.F = factors;
  //generate new variable factors
  get_vars(fg.V, factors);
  Factor* f;
  fg.F_v.clear();
  FOR1D(fg.V, i){
    f = new Factor(TUP(*fg.V(i)));
    f->setOne();
    fg.F_v.append(f);
  }
  //connect them with messages
  Variable *v;
  MessagePair* s;
  FOR1D(fg.F, i){
    f=fg.F(i);
    FOR1D(f->varIds, j){
      v = getvar(f->varIds(j));
      s = new MessagePair(f, v->factors.last());
      fg.msg_pairs.append(s);
    }
  }
  // beliefs
  fg.B_c.clear();
  fg.setCliqueBeliefs(factors);
  fg.B_v.clear();
  FOR1D(fg.V, i){
    f = new Factor(TUP(*fg.V(i)));
    f->setOne();
    fg.B_v.append(f);
  }

  fg.resetMessages();
  fg.setF2Bmap();
  fg.setV2F();
  fg.setV2Fv();
}
#else
void LoopyBP::constructBipartiteFactorGraph(FactorGraph& fg, const FacA& factors){
  uint i, k;
  // (1) variables
  get_vars(fg.V, factors);
  // (2) clique factors
  // original factors
  fg.F = factors;
  // beliefs
  fg.B_c.clear();
  fg.setCliqueBeliefs(factors);
  // (3) variable factors & (4) message pairs
  fg.F_v.clear();
  fg.B_v.clear();
  Factor* f;
  FOR1D(fg.V, i){
    f = new Factor(TUP(*fg.V(i)));
    f->setOne();
    fg.F_v.append(f);
    FOR1D(fg.F, k){
      if(fg.F(k)->varIds.findValue(*fg.V(i)) >= 0){
        MessagePair* s;
        if(fg.F(k)->varIds(0) == fg.V(i)->id)
          s = new MessagePair(fg.F(k), f);
        else
          s = new MessagePair(f, fg.F(k));
        fg.msg_pairs.append(s);
      }
    }
    // beliefs
    f = new Factor(TUP(*fg.V(i)));
    f->setOne();
    fg.B_v.append(f);
  }
  fg.resetMessages();
  fg.setF2Bmap();
  fg.setV2F();
  fg.setV2Fv();
}
#endif


//===========================================================================
//
// Algorithms on Factors and MessagePairs
//


//! collects a belief at a factor, optionally exluding one incoming message
void collectBelief(Factor& belief,const Factor& f,const MessagePair *exclude){
  MessagePair *s;
  uint i;
  belief = f;
  FOR1D(f.msg_pairs,i){
    s=f.msg_pairs(i);
    if(s==exclude) continue;
    if(s->f1==&f){
      tensorMultiply(belief,s->m21);
    }else{
      CHECK(s->f2==&f,"");
      tensorMultiply(belief,s->m12);
    }
  }
}

//! collects a belief at f1 and assigned m12 to its marginal
void recomputeMessage_12(MessagePair& sep){
  Factor belief;
  collectBelief(belief,*sep.f1,&sep);
  tensorMarginal(sep.m12,belief,sep.varIds);
}

//! collects a belief at f2 and assigned m21 to its marginal
void recomputeMessage_21(MessagePair& sep){
  Factor belief;
  collectBelief(belief,*sep.f2,&sep);
  tensorMarginal(sep.m21,belief,sep.varIds);
}

//! checks that marginals of connected factors are equal and the same as m12*m21
bool checkConsistency(const MessagePair& sep){
  Factor fmu1,fmu2,f1_marg,f2_marg,m12_m21;

  collectBelief(fmu1,*sep.f1,0);
  tensorMarginal(f1_marg, fmu1, sep.varIds);

  collectBelief(fmu2,*sep.f2,0);
  tensorMarginal(f2_marg, fmu2, sep.varIds);

  m12_m21=sep.m12;
  tensorMultiply(m12_m21,sep.m21);

  CHECK(f1_marg==f2_marg,"marginals inconsistent");
  CHECK(f1_marg==m12_m21,"marginals inconsistent");
  CHECK(m12_m21==f2_marg,"marginals inconsistent");
  return true;
}

//! checks a whole list of messages
bool checkConsistencyBatch(const MsgPairA& msgs){
  uint i;
  FOR1D(msgs, i) checkConsistency(*msgs(i));
  return true;
}


// Marc's eq (5) and (7)
void computeMessage_noDiv(Factor& f_from, Factor& f_to){
  uint DEBUG = 0;
  if(DEBUG>0){cout<<"computeMessage_noDivision [START]"<<endl;}
  uint i;
  FOR1D(f_from.msg_pairs, i){
    if(f_from.msg_pairs(i)->f1 == &f_to){
      recomputeMessage_21(*f_from.msg_pairs(i));
      break;
    }
    else if(f_from.msg_pairs(i)->f2 == &f_to){
      recomputeMessage_12(*f_from.msg_pairs(i));
      break;
    }
  }
  CHECK(f_from.msg_pairs.N > i, "Factors don't have a MessagePair!");
  if(DEBUG>0){cout<<"computeMessage_noDivision [END]"<<endl;}
}


void computeMessage_withDiv(Factor& f_from, Factor& f_to){
  uint DEBUG = 0;
  if(DEBUG>0){cout<<"calcMessage [START]"<<endl;}
  // find MessagePair
  MessagePair* s = getMessagePair(&f_from, &f_to);
  if(&f_from == s->f2){
    if(DEBUG > 0){cout <<f_from.varIds <<" --> " <<s->f1->varIds<<endl;}
    if(DEBUG > 2){
      cout <<"Message 2 --> 1"<<endl;
      cout <<"f_from: " <<endl <<f_from;
    }
    Factor belief;
    collectBelief(belief, f_from, 0);
    tensorMarginal(s->m21, belief, s->varIds);
    if(DEBUG>2){
      cout <<"Belief over f_from: " <<endl <<belief;
      cout <<"Old msg in other direction (m12): " <<endl <<s->m12;
    }
    tensorDivide(s->m21, s->m12);
    if(DEBUG > 1) cout <<"New msg (m21): " <<endl <<s->m21;
  }
  // message m12
  else {
    if(DEBUG > 0){cout <<f_from.varIds <<" --> " <<s->f2->varIds<<endl;}
    if(DEBUG > 2){
      cout <<"Message 1 --> 2"<<endl;
      cout <<"f_from: " <<endl <<f_from <<endl;
    }
    Factor belief;
    collectBelief(belief, f_from, 0);
    tensorMarginal(s->m12, belief, s->varIds);
    if(DEBUG>2){
      cout <<"Belief over f_from: " <<endl <<belief;
      cout <<"Old msg in other direction (m21): " <<endl <<s->m21;
    }
    tensorDivide(s->m12, s->m21);
    if(DEBUG > 1) cout <<"New msg (m12): " <<endl <<s->m12;
  }
  if(DEBUG>0){cout<<"lbp_calcMessage [END]"<<endl;}
}


void computeMessage_noDiv(Factor& f_from, Factor& b_from, Factor& f_to){
  HALT("this is actually with div...");
  uint DEBUG = 0;
  if(DEBUG>0){cout<<"calcMessage [START]"<<endl;}
  // find MessagePair
  MessagePair* s = getMessagePair(&f_from, &f_to);
  if(&f_from == s->f2){
    if(DEBUG > 0){cout <<f_from.varIds <<" --> " <<s->f1->varIds<<endl;}
    if(DEBUG > 2){
      cout <<"Message 2 --> 1"<<endl;
      cout <<"f_from: " <<endl <<f_from;
    }
    tensorMarginal(s->m21, b_from, s->varIds);
    if(DEBUG>2){
      cout <<"Marginalised f_from: " <<endl <<s->m21;
      cout <<"Old msg in other direction (m12): " <<endl <<s->m12;
    }
    tensorDivide(s->m21, s->m12);
    if(DEBUG > 1) cout <<"New msg (m21): " <<endl <<s->m21;
  }
  // message m12
  else {
    if(DEBUG > 0){cout <<f_from.varIds <<" --> " <<s->f2->varIds<<endl;}
    if(DEBUG > 2){
      cout <<"Message 1 --> 2"<<endl;
      cout <<"f_from: " <<endl <<f_from <<endl;
    }
    tensorMarginal(s->m12, b_from, s->varIds);
    if(DEBUG>2){
      cout <<"Marginalised f_from: " <<endl <<s->m12;
      cout <<"Old msg in other direction (m21): " <<endl <<s->m21;
    }
    tensorDivide(s->m12, s->m21);
    if(DEBUG > 1) cout <<"New msg (m12): " <<endl <<s->m12;
  }
  if(DEBUG>0){cout<<"lbp_calcMessage [END]"<<endl;}
}






double passMessage(Factor& f_from, Factor& f_to, Factor& b_to, MsgCalc calcMsgType){
  uint DEBUG = 0;
  
  // store old
  // delete this later (for efficiency reasons)
  Factor belief_to_old;
  collectBelief(belief_to_old, f_to, 0); 
  arr p_old = belief_to_old.P;
  
  if(DEBUG>0){
    cout<<"passMessage: "<<endl;
    cout<<"f_from: "<<f_from<<endl;
    cout<<"f_to: "<<f_to<<endl;
  }
  
  // (1) calc new message
  if(calcMsgType == WITH_DIV)
    computeMessage_withDiv(f_from, f_to);
  else
    computeMessage_noDiv(f_from, f_to);
  // (2) recalculate belief accordingly
  collectBelief(b_to, f_to, 0);
  
  
  if(DEBUG>0){
    cout<<"Message:"<<endl;
    writeMessage(&f_from, &f_to, cout);cout<<endl;
    cout<<"Updated b_to:"<<endl<<b_to<<endl;
    cout<<" --> "<<(b_to.P - p_old).absMax()<<endl;
  }
  
  // calc change (and ignore log_P !)
  return (b_to.P - p_old).absMax();
}



// double distributeMessages(FactorGraph& fg, Factor& f, MsgCalc calcMsgType){
//   uint DEBUG = 0;
//   uint i, index;
//   double change, max_change=0.0;
//   if(DEBUG>0){cout<<"Distributing messages from factor over "<<f.varIds<<endl;}
//   if(DEBUG>1){cout<<f<<endl;}
//   FOR1D(f.msg_pairs, i){
//     Factor* neighbor;
//     Factor* neighbor_belief;
//     if(f.msg_pairs(i)->f1 == &f){
//       index = get_list_index_unsigned(fg, f.msg_pairs(i)->f2);
//       neighbor = fg.
//       if(DEBUG>1){cout<<"to (before) "; f.msg_pairs(i)->f2->writeExtremelyNice(cout);cout<<endl;}
//       change = passMessage_1(f, *f.msg_pairs(i)->f2, calcMsgType); // based on Marc's eq (2) / (3)
//       if(max_change<change) max_change=change;
//       if(DEBUG>1){cout<<"to (after) "; f.msg_pairs(i)->f2->writeExtremelyNice(cout);cout<<endl;}
//     }
//     else {
//       if(DEBUG>1){cout<<"to (before) "; f.msg_pairs(i)->f1->writeExtremelyNice(cout);cout<<endl;}
//       change = fg.passMessage_1(f, *f.msg_pairs(i)->f1, calcMsgType); // based on Marc's eq (2) / (3)
//       if(max_change<change) max_change=change;
//       if(DEBUG>1){cout<<"to (after) "; f.msg_pairs(i)->f1->writeExtremelyNice(cout);cout<<endl;}
//     }
//   }
//   return max_change;
// }


void getMarginal(Factor& marginal, const uintA& marginalVars, FactorGraph& fg){
  // search for factors containing marginalVars
  uint f, v;
  FOR1D(fg.B_c, f){
    FOR1D(marginalVars, v){
      if(fg.B_c(f)->varIds.findValue(marginalVars(v))<0)
        break;
    }
    if(marginalVars.N > v)
      continue;
    else {
      tensorMarginal(marginal, *fg.B_c(f), marginalVars);
      return;
    }
  }
  HALT("No factor with all marginalVars " <<marginalVars <<" -- " <<getvar(marginalVars(0))->name);
}


/* return a list of posteriors for each variable.
  We step through all variables, get the (last) factor over
  this variable, collect a belief for it, and marginalize it */
void getVariableBeliefs(MT::Array<arr>& post,const VarA& vars){
  uint i,N=vars.N;
  post.resize(N);
  Factor *f,belief,marg;
  for(i=0;i<N;i++){
    f=vars(i)->factors.last(); //simply take last factor in factor list!
    collectBelief(belief,*f,NULL);
    tensorMarginal(marg,belief,TUP(vars(i)->id));
    post(i) = marg.P;
  }
}






//===========================================================================
//
// convenience methods
//
//

iSpace* get_global_space(){
  return &globalInferSpace;
}

Variable* getvar(uint id){
//   MT_MSG("not always is the global space the only one!!");
  return get_global_space()->variables(id);
}

void get_global_vars(MT::Array<Variable*>& vars){
  vars.clear();
  vars = get_global_space()->variables;
}

void printvars(uintA ids){
  uint i;
  FOR1D(ids, i){
    cout<<getvar(ids(i))->name<<"  ";
  }
}

void printFactorArrayViaVariableSets(FacA factors){
  uint i;
  FOR1D(factors, i){
    cout <<factors(i)->varIds <<" ";
  }
  cout <<endl;
}




//===========================================================================
//
// write methods
//
//

void writeNice(const VarA& vars){
  uint i;
  FOR1D(vars, i){
    vars(i)->write(cout);cout<<endl;
  }
}

void writeEdges(const FacA& individual_factors, bool withProbs){
  uint i, k;
  FOR1D(individual_factors, i){
    FOR1D(individual_factors(i)->varIds, k){
      if(k==0) continue;
      cout<<get_global_space()->variables(individual_factors(i)->varIds(k))->name<<"  ";
    }
    cout<<"  --->   "<<get_global_space()->variables(individual_factors(i)->varIds(0))->name<<endl;
    //         }
    if(withProbs){
      cout <<" P" <<individual_factors(i)->P;
      cout <<'*' <<::exp(individual_factors(i)->logP) <<"=exp(" <<individual_factors(i)->logP <<") " <<endl;
    }
  }
}


void write(const FacA& facs, ostream& os){
  uint i;
  FOR1D(facs, i){
    os<<"["<<i<<"]  ";
    os<<*facs(i)<<endl;
  }
}

void writeNice(const FacA& individual_factors){
  uint i, k;
  FOR1D(individual_factors, i){
    FOR1D(individual_factors(i)->varIds, k){
      cout<<get_global_space()->variables(individual_factors(i)->varIds(k))->name<<"  ";
    }
    cout<<endl;
  }
}

void writeExtremelyNice(const FacA& facs){
  uint i;
  FOR1D(facs, i){
    facs(i)->writeNice();cout<<endl;
    facs(i)->writeExtremelyNice();cout<<endl;
  }
}










//===========================================================================
//  
//  Factor Algebra

double robustDivide(double a,double b){
  if(!a) if(!b) return 1.; else return 0.;
  //if(a==b) return 1.;
  if(b) return a/b;
  HALT("division by zero: " <<a <<"/" <<b);
  return 0;
}

/*!\brief If id is a tuple of k variable identifiers, and mid is a _subset_ of l of those,
then pick will contain l indices, shouting which slot of id corresponds to each slot of mid. For instance
if we have a product of two sensors \f$A_{ijklm} B_{kli}\f$ then pick will be <2 3 0> shouting that
the 0th slot of B is the 2nd of A, that the 1st slot of B is the 3rd of A, and that the 3rd slot of B is the 0th of A. */
void getPick(uintA& pick,const uintA& id,const uintA& mid){
  uint i=0,k=0;
  pick.resize(mid.N);
  for(k=0;k<mid.N;k++){
    for(i=0;i<id.N;i++) if(id(i)==mid(k)) break;
    CHECK(i<id.N,"getPick error: id="<<id <<" mid=" <<mid);
    pick(k)=i;
  }
  DEBUG_INFER(3,cout <<"** getPick: id=" <<id <<" mid=" <<mid <<" pick=" <<pick <<endl);
}

void tensorProduct(Factor& f, Factor& a, Factor& b){
  CHECK(a.space==b.space,"");
  uintA fid;
  setUnite(fid,a.varIds,b.varIds);
  f.init(a.space,fid);
  f.P.resize(f.dim);
  uintA pickA,pickB;
  getPick(pickA,f.varIds,a.varIds);
  getPick(pickB,f.varIds,b.varIds);
  tensorEquation(f.P, a.P, pickA, b.P, pickB,0);
  f.logP = a.logP + b.logP;
  lognormScale(f.P,f.logP);
}

void tensorProductMarginal(Factor& f, const Factor& a, const Factor& b, const uintA& s){
  CHECK(a.space==b.space,"");
  uintA fid;
  setUnite(fid,a.varIds,b.varIds);
  setMinus(fid,s);
  f.init(a.space,fid);
  f.P.resize(f.dim);
  uintA all,pickA,pickB;
  all=f.varIds;
  all.append(s);
  getPick(pickA,all,a.varIds);
  getPick(pickB,all,b.varIds);
  tensorEquation(f.P, a.P, pickA, b.P, pickB, s.N);
  f.logP = a.logP + b.logP;
  lognormScale(f.P,f.logP);
}

void tensorMarginal(Factor& m,const Factor& f, const uintA& marginalVars){
  m.init(f.space,marginalVars);
  uintA pick;
  getPick(pick,f.varIds,m.varIds);
  tensorMarginal(m.P,f.P,pick);
  m.logP=f.logP;
  lognormScale(m.P,m.logP);
}

void tensorMaxMarginal(Factor& m,const Factor& f, const uintA& marginalVars){
  m.init(f.space,marginalVars);
  uintA pick;
  getPick(pick,f.varIds,m.varIds);
  tensorMaxMarginal(m.P,f.P,pick);
  //do we need to normalize in the max-marginal case?
  //normalizeDist(Y);
  m.logP=f.logP;
  lognormScale(m.P,m.logP);
}

void tensorMultiply(Factor& f,const Factor& m){
  CHECK(f.space==m.space,"");
  if(m.varIds==f.varIds){
    mult(f.P,f.P,m.P);
  }else{
    uintA pick;
    getPick(pick,f.varIds,m.varIds);
    tensorMultiply(f.P,m.P,pick);
  }
  f.logP += m.logP;
  lognormScale(f.P,f.logP);
}

void tensorDivide(Factor& f,const Factor& m){
  CHECK(f.space==m.space,"tensorDivide: factors from different spaces");
  if(m.varIds==f.varIds){
    div(f.P,f.P,m.P);
  }else{
    uintA pick;
    getPick(pick,f.varIds,m.varIds);
    tensorDivide(f.P,m.P,pick);
  }
  f.logP -= m.logP;
  //   cout <<"infer.cpp.tensorDivide vormScalen: " <<f.P <<endl;
  lognormScale(f.P,f.logP);
  //   cout <<"infer.cpp.tensorDivide nachmScalen: " <<f.P <<endl;
}

void tensorInvertMultiply(Factor& f,const Factor& m){
  CHECK(f.space==m.space,"");
  CHECK(m.varIds==f.varIds,"Factor invMultiply needs identical variables");
  for(uint i=0;i<f.P.N;i++) f.P.elem(i)=robustDivide(m.P.elem(i),f.P.elem(i));
  f.logP = -f.logP + m.logP;
  lognormScale(f.P,f.logP);
}

void tensorWeightedAdd(Factor& f,double w,const Factor& m){
  CHECK(f.space==m.space,"");
  CHECK(m.varIds==f.varIds,"Factor weightedAdd needs identical variables");
  w *= ::exp(m.logP-f.logP);
  for(uint i=0;i<f.P.N;i++) f.P.elem(i) = f.P.elem(i) + w*m.P.elem(i);
  lognormScale(f.P,f.logP);
}

void writeInfo(ostream& os,const Factor& f){
  uint i;
  os <<"vars=(";
  for(i=0;i<f.varIds.N;i++){
    if(i) os <<' ';
    os <<f.varIds(i) <<'.' <<f.space->variables(f.varIds(i))->name;
    CHECK(f.varIds(i)==f.space->variables(f.varIds(i))->id,"identity mismatch!!");
  }
  os <<") dims=<";
  for(i=0;i<f.varIds.N;i++){
    if(i) os <<' ';
    os <<f.dim(i);
    CHECK(f.dim(i)==f.space->variables(f.varIds(i))->dim,"dimensionality mismatch!!");
  }
  os <<"> P.dims=";
  f.P.writeDim(os);
  os <<endl;
}




//===========================================================================
//
// ELIMINATION ALGORITHM methods
//
void getJoint(Factor& joint,const FacA& factors){
  DEBUG_INFER(1,cout <<MT_HERE <<endl);
  uint i;
  //get tuple of vars
  uintA jointVars;
  for(i=0;i<factors.N;i++) jointVars.setAppend(factors(i)->varIds);
  DEBUG_INFER(2,cout <<"  jointVars=" <<jointVars <<endl);
  //compute joint
  joint.init(factors(0)->space,jointVars);
  joint.setOne();
  for(i=0;i<factors.N;i++) tensorMultiply(joint, *(factors(i)));
}

/*! Computes an order for the elimination of variables.
The order is such that in each iteration among all variables that have not been eliminated yet 
the one which has the fewest links to other not-deleted variables is chosen.
This is equivalent to saying that per iteration the variable chosen for elimination is the one
that would create the smallest clique if all factors that involve this variable were multiplied.
*/
void computeEliminationOrder(uintA& elimOrder, const FacA& factors, const uintA& elimVars){
  int DEBUG_INFER_LEVEL = 0;
  DEBUG_INFER(1,cout <<MT_HERE <<endl);
  DEBUG_INFER(2,cout <<"  input factors=\n" <<factors <<endl);
  DEBUG_INFER(1,cout <<"variables to eliminate=" <<elimVars <<endl);
  
  elimOrder.resize(elimVars.N);
  uint f, v, e;
  iSpace *space = &globalInferSpace;
  
  // Determine for each variable the set of variables it is linked to.
  MT::Array<uintA> connectedVarSets(elimVars.N);
  for(v=0; v<elimVars.N; v++){
    for(f=0; f<factors.N; f++){
      if(factors(f)->varIds.findValue(elimVars(v))>-1){
        connectedVarSets(v).setAppend(factors(f)->varIds);
      }
    }
    DEBUG_INFER(2,cout <<"  neighbors of v"<<elimVars(v) <<"=" <<connectedVarSets(v) <<endl);
  }
  // Calculate elimination order
  boolA used(elimOrder.N);
  used=false;
  DEBUG_INFER(1,cout <<"calculation of elimination order: " <<endl);
  uint cost, minCost=-1;
  int minIndex;
  for(e=0; e<elimOrder.N; e++){
    DEBUG_INFER(2, cout <<"  (" <<e <<")" <<endl);
    // Determine variable with smallest set of connected remaining variables
    minIndex=-1;
    for(v=0; v<elimVars.N; v++){
      if(used(v)) continue;  //don't consider already used vars
#if 1
      //determine cost as the #variables in the generated clique
      cost=connectedVarSets(v).N;
#else
      //determine cost as the size (=size of probability table) of the generated clique
      cost=1;
      for(uint i=0;i<connectedVarSets(v).N;i++)
        if(connectedVarSets(v)(i)!=v)  cost*=space->variables(connectedVarSets(v)(i))->dim;
#endif      
      DEBUG_INFER(2, cout <<"  testing v" <<elimVars(v) <<": cost=" <<cost <<endl);
      if(minIndex==-1 || minCost>cost){
        minCost = cost;
        minIndex = v;
      }
    }
    elimOrder(e) = elimVars(minIndex);
    used(minIndex) = true;
    DEBUG_INFER(1, cout <<"eliminating v" <<elimOrder(e) <<" '" <<space->variables(elimOrder(e))->name <<"' (cost=" <<minCost<<")" <<endl);
    // Delete eliminated variable out of sets of connected variables of other vars
    for(v=0; v<elimVars.N; v++){
      connectedVarSets(v).removeValueSafe(elimOrder(e));
      DEBUG_INFER(2, cout <<"  neighbors of v" <<elimVars(v) <<"=" <<connectedVarSets(v) <<endl);
    }
  }
  DEBUG_INFER(1, cout <<"elimination order: " <<elimOrder <<endl);
}

/*! eliminates a single variable from the factor list.

At input, factors describes the full model; at output, factors contains the reduced model which include unchanged old factors and some newed factors.
The newed factors are additionally appended to the newed_factors (to allow for external cleanup) */
void eliminateVariable(FacA& factors, FacA& newed_factors,uint var){
  uint f;
  FacA referencedFactors;
  factors.memMove=true;
  referencedFactors.memMove=true;
  //collect referenced factors
  for(f=0;f<factors.N;f++)
    if(factors(f)->varIds.findValue(var)!=-1) referencedFactors.append(factors(f));

  //compute joint variable tuple
  uintA jointVars;
  for(f=0;f<referencedFactors.N;f++)
    jointVars.setAppend(referencedFactors(f)->varIds);
    //setUnite(jointVars,jointVars,referencedFactors(f)->varIds);//[mt]

  //compute joint clique and remove! the used factors from the list
  Factor jointFactor(jointVars);
  //jointFactor.setOne(); is done in the constructor
  for(f=0;f<referencedFactors.d0; f++){
    tensorMultiply(jointFactor, *(referencedFactors(f)));
    factors.removeValue(referencedFactors(f));
  }

  //eliminate the single variable from the joint factor
  jointVars.removeValue(var);
  Factor *newFactor = new Factor(jointVars);
  tensorMarginal(*newFactor,jointFactor, jointVars);
  factors.append(newFactor);
  newed_factors.append(newFactor);
  
  //DEBUG_INFER(1,checkSpaceRegistry());
}

/*! marginalizes a factor list over all variables except the "remaining_vars". The output is a
single factor over the remaining_vars with the marginal. The factors list remains unchanged. */
void eliminationAlgorithm(Factor& posterior,const FacA& factors, const uintA& remaining_vars){
  DEBUG_INFER(1,cout <<MT_HERE <<endl);
  uint i,f;
  // determine which variables need to be eliminated: ALL \ post_vars
  uintA elimVars;
  for(f=0;f<factors.N;f++)        elimVars.setAppend(factors(f)->varIds);
  
  DEBUG_INFER(2,cout <<"  all  vars=" <<elimVars <<"remaining vars=" <<remaining_vars <<endl);
  
  for(i=0;i<remaining_vars.N;i++) elimVars.removeValue(remaining_vars(i));

  DEBUG_INFER(2,cout <<"  elim vars=" <<elimVars <<endl);
  
  // determine order in which variables are eliminated.
  uintA elimOrder;
  computeEliminationOrder(elimOrder, factors, elimVars);

  DEBUG_INFER(2,cout <<"  elimination order=" <<elimOrder <<endl);
  
  // eliminate in this order
  FacA factors_copy(factors);
  FacA newedFactors;
  for(i=0;i<elimOrder.N;i++) eliminateVariable(factors_copy, newedFactors, elimOrder(i));

  // calculate posterior
  posterior.init(0,remaining_vars);
  posterior.setOne();
  for(i=0;i<factors_copy.N;i++){
    DEBUG_INFER(2,cout <<"  remaining factor " <<i<<"=" <<*factors_copy(i) <<endl);
    tensorMultiply(posterior,*factors_copy(i));
  }
  
  DEBUG_INFER(2,cout <<"  posterior=" <<posterior <<endl);

  //cleanup newed factores
  for(i=0;i<newedFactors.N;i++) delete newedFactors(i);
}


void moralize(FacA& factorsOfDirectedGraph, FacA& factorsOfMoralizedGraph){
  // first variable is child, remaining variables are parents
  NIY;
}


//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//===========================================================================
//
// JUNCTION TREE methods
//

void checkJunctionTreeProperty_dfs(Factor* node, Factor* parent, FactorGraph& junctionTree, boolA& containsId, uint id){
  int DEBUG = 0;
  uint i;
  containsId(get_list_index_unsigned__orig(junctionTree, node)) = 0;
  FacA neighbors;
  getNeighbors(node, neighbors);
  FOR1D(neighbors, i){
    if(neighbors(i) == parent)
      continue;
    if(DEBUG > 0)
      cout <<"Checking direction " <<node->varIds <<" --> " <<neighbors(i)->varIds <<endl;
    // check whether neighbor contains varible
    if(neighbors(i)->varIds.findValue(id) != -1){
      if(DEBUG > 0)
        cout <<"   --> going on that way" <<endl;
      // check that variable contained in MessagePair
#ifndef MT_NOCHECK
      MessagePair* s = getMessagePair(node, neighbors(i));
#endif
      CHECK(s->varIds.findValue(id) != -1, "Variable " <<id <<" is not contained in MessagePair " <<s->varIds <<" for factors over " <<node->varIds <<" and " <<neighbors(i)->varIds <<"!" <<endl);
      // continue dfs here
      checkJunctionTreeProperty_dfs(neighbors(i), node, junctionTree, containsId, id);
    }
    else {
      if(DEBUG > 0)
        cout <<"   --> dead end!" <<endl;
      continue;
    }
  }
}

// Jordan, Chapter 17, p.22: For every pair of cliques V and W, all cliques on (unique) path between V and W contain section(V,W).
// --> checks that each variable induces a single subtree in junction tree
void JunctionTree::checkJunctionTreeProperty(FactorGraph& junctionTree){
  int DEBUG = 0;
  if(DEBUG > 0)
    cout <<"checkJunctionTreeProperty [START]" <<endl;
  
  uint v, f;
  uint id;
  boolA containsId(junctionTree.B_c.N);
  FOR1D(junctionTree.V, v){
    id = junctionTree.V(v)->id;
    if(DEBUG > 0)
      cout <<" id=" <<id <<endl;
    FOR1D(junctionTree.B_c, f){
      if(junctionTree.B_c(f)->varIds.findValue(id) != -1)
        containsId(f) = 1;
      else
        containsId(f) = 0;
    }
    if(DEBUG > 0)
      cout <<" containsId_before=" <<containsId <<endl;
    f=0;
    while(f<junctionTree.B_c.N && !containsId(f))
      f++;
    CHECK(f<junctionTree.B_c.N, " Variable ID " <<id <<" has not been found in graph." <<endl);
    checkJunctionTreeProperty_dfs(junctionTree.B_c(f), 0, junctionTree, containsId, id);
    if(DEBUG > 0)
      cout <<" containsId_after=" <<containsId <<endl;
    FOR1D(containsId, f)
      CHECK(containsId(f) == 0, " junction tree property violated for variable " <<id <<endl);
  }
  if(DEBUG > 0)
    cout <<"checkJunctionTreeProperty [END]" <<endl;
}



void JunctionTree::addEvidence(FactorGraph& junctionTree, Factor& evid){
  uint f;
  uintA varSection;
  FOR1D(junctionTree.F, f){
    setSection(varSection, evid.varIds, junctionTree.F(f)->varIds);
    if(varSection.N > 0){
      MessagePair* s = new MessagePair(junctionTree.F(f), &evid);
      junctionTree.F.append(&evid);
      // update orig factor
      junctionTree.B_c.append(new Factor());
      *junctionTree.B_c.last() = evid;
      junctionTree.msg_pairs.append(s);
      break;
    }
  }
}





void JunctionTree::buildTriangulatedCliques(const FacA& factors, FacA& triangulatedCliques){
  bool DEBUG = false;
  bool DEBUG_VERBOSE = false;
  
  uint v, v2, v3, f; 
  FacA intermediateFactors = factors;
  boolA intermediateFactors_removed(intermediateFactors.N);
  FOR1D(intermediateFactors_removed, v)
    intermediateFactors_removed(v) = 0;
  
  // determine existing variables
  uintA vars;
  for(f=0; f < factors.N; f++){
    setUnite(vars, vars, factors(f)->varIds);
  }
  std::sort(vars.p, vars.p + vars.N);
  if(DEBUG)
    cout <<"Existing vars: " <<vars <<endl;
  
  
  // map varID -> order over participating vars
  std::map<uint, uint> var_id2order;
  FOR1D(vars, v){
    var_id2order[vars(v)] = v;
  }
  if(DEBUG){
    std::map<uint, uint>::iterator itVars;
    for( itVars = var_id2order.begin() ; itVars != var_id2order.end(); itVars++ ){
      cout <<"var_id2order[" <<itVars->first <<"] = " <<itVars->second <<endl;
    }
  }
  
  // determine elimination ordering
  // 	uintA elimOrder(6);
  // 	for(v=0; v<6; v++){
  // 		elimOrder(v) = 5-v;
  // 	}
  uintA elimOrder;
  computeEliminationOrder(elimOrder, intermediateFactors, vars);
  //     elimOrder <<"[   9   14   7   8   10   11   12   13   0    1    2    15   16   3   4   5   6]";
  //     elimOrder <<"[   12   17   10   11   13   14   15   16  7   0   1   2   8   9   3   4   5   6 ]";
  //     elimOrder <<"[   19   29   16   17   18   20   21   22   23   24   25   26   27   28   0   12   1   2   13   14   15   3   4   5   6   7   8   9   10   11]";
  //     elimOrder = global_elimOrder;
  
  if(DEBUG){
    cout <<"Elimination order: " <<elimOrder <<endl;
    if(DEBUG_VERBOSE){
      FOR1D(elimOrder, v){
        cout<<getvar(elimOrder(v))->name<<"  ";
      }
      cout<<endl;
    }
  }
  
  // edges matrix
  // indexed by var ids
  MT::Array<bool> edges(elimOrder.N, elimOrder.N);
  FOR2D(edges,v,v2){
    edges(v,v2) = 0;
  }
  // init edges matrix with existing edges in original graph
  FOR1D(intermediateFactors, f){
    FOR1D(intermediateFactors(f)->varIds, v){
      for(v2 = v+1; v2 < intermediateFactors(f)->varIds.N; v2++){
        edges(var_id2order[intermediateFactors(f)->varIds(v)], var_id2order[intermediateFactors(f)->varIds(v2)]) = 1;
        edges(var_id2order[intermediateFactors(f)->varIds(v2)], var_id2order[intermediateFactors(f)->varIds(v)]) = 1;
      }
    }
  }
  // 	if(DEBUG)
  // 		cout <<"Edges matrix:" <<endl <<edges <<endl;
  
  
  // UNDIRECTED_GRAPH_ELIMINATE
  
  FOR1D(elimOrder, v){
    if(DEBUG){
      cout <<"--------------------------------------------------" <<endl;
      cout <<"(" <<v <<")" <<" Eliminating " <<elimOrder(v) <<endl;
      if(DEBUG_VERBOSE){cout <<"(" <<v <<")" <<" Eliminating " <<getvar(elimOrder(v))->name <<endl;}
    }
    // determine remaining neighbors
    uintA remainingNeighbors;
    FOR1D(edges, v2){
      if(edges(var_id2order[elimOrder(v)], var_id2order[elimOrder(v2)]))
        remainingNeighbors.append(elimOrder(v2));
    }
    if(DEBUG){
      cout <<"Remaining neighbors:" <<remainingNeighbors <<endl;
      if(DEBUG_VERBOSE){cout <<"Remaining neighbors: "; printvars(remainingNeighbors); cout<<endl;}
    }
    
    // connect remaining neighbors
    FOR1D(remainingNeighbors, v2){
      for(v3 = v2+1; v3<remainingNeighbors.N; v3++){
        edges(var_id2order[remainingNeighbors(v2)], var_id2order[remainingNeighbors(v3)]) = 1;
        edges(var_id2order[remainingNeighbors(v3)], var_id2order[remainingNeighbors(v2)]) = 1;
      }
    }
    // 		if(DEBUG)
    // 			cout <<"Edges matrix (remaining neighbors connected):" <<endl <<edges <<endl;
    
    // set up new factor with all neighbors
    if(DEBUG)
      cout <<"Setting up new factor [START]" <<endl;
    uintA vars;
    vars.append(elimOrder(v));
    setUnite(vars, vars, remainingNeighbors);
    if(DEBUG)
      cout <<"Variables ["<<vars.N<<"]: " <<vars <<endl;
    if(DEBUG_VERBOSE){cout <<"Variables: "; printvars(vars);cout<<endl;}
    // check whether "vars" of new factors are subset of existing factor f
    bool containedInOtherFactor = 0;
    FOR1D(triangulatedCliques, f){
      if(numberSharedElements(triangulatedCliques(f)->varIds, vars) == vars.N){
        containedInOtherFactor = true;
        break;
      }
    }
    Factor* newFactor;
    // if so, include into f
    if(containedInOtherFactor){
      newFactor = triangulatedCliques(f);
      if(DEBUG)
        cout <<"--> included in " <<newFactor->varIds <<endl;
      if(DEBUG_VERBOSE){cout <<"--> included in "; printvars(newFactor->varIds);cout<<endl;}
    }
    // else construct new factor
    else {
      if(DEBUG)
        cout <<"Spaeter: Variables ["<<vars.N<<"]: " <<vars <<endl;
      // TODO weg
      //             if(vars.N == 28) continue;
      
      newFactor = new Factor(vars);
      newFactor->setOne();
      // add new factor to triangulated set
      triangulatedCliques.append(newFactor);
      if(DEBUG)
        cout <<"--> true new factor" <<endl;
    }
    
    if(DEBUG)
      cout <<"Calculating new potential:" <<endl;
    // calc new potential
    for(f=0; f < intermediateFactors.N; f++){
      if(intermediateFactors_removed(f) != 1
    				&& intermediateFactors(f)->varIds.findValue(elimOrder(v)) > -1){
        if(DEBUG)
          cout <<" ** Using factor: " <<*intermediateFactors(f);
        tensorMultiply(*newFactor, *(intermediateFactors(f)));
        if(DEBUG)
          cout <<"  --> updated newFactor: " <<*newFactor;
        intermediateFactors_removed(f) = 1;
      }
      if(DEBUG){
        cout<<"xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"<<endl;
        cout<<"All factors thus far:"<<endl;
        FOR1D(triangulatedCliques,v2){
          cout<<v2<<":  "<<triangulatedCliques(v2)->varIds.N<<endl;
        }
      }
    }
    if(DEBUG){
      cout <<"Final newFactor: " <<endl <<*newFactor;
      if(DEBUG_VERBOSE){cout <<" over vars "; printvars(newFactor->varIds);cout<<endl;}
      cout <<"Setting up new factor [END]" <<endl;
    }
    
    // remove node from graph
    for(v2=0; v2 < elimOrder.N; v2++){
      edges(var_id2order[elimOrder(v)], var_id2order[elimOrder(v2)]) = 0;
      edges(var_id2order[elimOrder(v2)], var_id2order[elimOrder(v)]) = 0;
    }
    
    if(DEBUG){
      // 			cout <<"Edges matrix (node removed):" <<endl <<edges <<endl;
      cout <<"Removed factors: " <<intermediateFactors_removed <<endl;
    }
        }
}


// Kruskal algorithm
void JunctionTree::buildMaxSpanningTree(FacA& factors, FactorGraph& cliqueTree){
  uint DEBUG = 0;
  if(DEBUG >= 1){
    cout <<"========================================" <<endl;
    cout <<"buildMaxSpanningTree" <<endl;
    cout <<"--------------------" <<endl;
  }
  
  uint f, f2, maxId;
  uint max;
  Factor *fac1, *fac2;
  uintA tempA;
  
  if(DEBUG >= 1){
    cout <<"input factors [START]" <<endl;
    FOR1D(factors, f){
      cout <<"(" <<f <<") " <<factors(f)->varIds <<endl;
    }
    cout <<"input factors [END]" <<endl;
  }
  
  // check that each factor has a non-empty MessagePair for some other factor
  // = shares variables with some other factor
  // = not disconnected from tree
  boolA sharesVariables(factors.N);
  FOR1D(sharesVariables, f)
    sharesVariables(f) = false;
  FOR1D(factors, f){
    if(sharesVariables(f))
      continue;
    FOR1D(factors, f2){
      if(f == f2)
        continue;
      setSection(tempA, factors(f)->varIds, factors(f2)->varIds);
      if(tempA.N > 0){
        sharesVariables(f) = true;
        sharesVariables(f2) = true;
        break;
      }
    }
    // might be a single factor that contains all vars
    if(!sharesVariables(f)){
      FOR1D(factors(f)->space->variables, f2){
        if(factors(f)->varIds.findValue(*factors(f)->space->variables(f2)) < 0)
          break;
      }
      if(f2 == factors(f)->varIds.N)
        sharesVariables(f) = true;
    }
    if(!sharesVariables(f))
      HALT("While building max spanning tree: factor node over " <<factors(f)->varIds <<" is disconnected from junction tree!");
  }
  
  
  MT::Array<MessagePair*> candidate_msg_pairs;
  boolA candidate_msg_pairs_contained;
  FOR1D(factors, f){
    for(f2 = f+1; f2 < factors.N; f2++){
      setSection(tempA, factors(f)->varIds, factors(f2)->varIds);
      if(tempA.N > 0){
        MessagePair* s = new MessagePair(factors(f), factors(f2));
        candidate_msg_pairs.append(s);
        candidate_msg_pairs_contained.append(1);
      }
    }
  }
  
  if(DEBUG >= 1){
    cout <<"candidate_msg_pairs [START]" <<endl;
    FOR1D(candidate_msg_pairs, f){
      cout <<*(candidate_msg_pairs(f)) <<endl;
    }
    cout <<"candidate_msg_pairs [END]" <<endl;
  }
  
  // define elementary clusters
  // factorID --> Cluster (made of factor IDs)
  MT::Array< uintA > clusters(factors.N);
  FOR1D(factors, f){clusters(f).append(f);}
  if(DEBUG >= 1){
    cout <<"Elementary clusters [START]" <<endl;
    FOR1D(factors, f){
      cout <<"(" <<f <<") Factor over " <<factors(f)->varIds <<" connected to:" <<endl;
      FOR1D(clusters(f), f2){
        cout <<factors((clusters(f))(f2))->varIds <<"  ";
      }
      cout<<endl;
      // ids only:
      cout<<"In ids: #"<<f<<" -> "<<clusters(f)<<endl;
    }
    cout <<"Elementary clusters [END]" <<endl;
  }
  
  // THE ALGORITHM
  // big loop	
  if(DEBUG >= 1){
    cout <<endl <<"---------------------------------------" <<endl;
    cout <<"THE ALGORITHM [START]" <<endl <<endl;
  }
  while(cliqueTree.msg_pairs.N < factors.N -1){
    // determine S with max weight
    maxId = 1000;
    max = 0;
    FOR1D(candidate_msg_pairs, f){
      if(candidate_msg_pairs_contained(f)
        && candidate_msg_pairs(f)->varIds.N > max){
        maxId = f;
        max = candidate_msg_pairs(f)->varIds.N;
      }
    }
    if(maxId == 1000){
      HALT("Constructing max spanning tree failed: No more MessagePair candidates with |MessagePair set| > 0 found!");
    }
    
    if(DEBUG >= 1){
      cout <<"++++++++++++++" <<endl;
      cout <<"Next MessagePair candidate:" <<endl;
      cout <<*(candidate_msg_pairs(maxId));
    }
    fac1 = candidate_msg_pairs(maxId)->f1;
    fac2 = candidate_msg_pairs(maxId)->f2;
    // index of fac1 in factors
    f = factors.findValue(fac1);
    // index of fac2 in factors
    f2 = factors.findValue(fac2);
    // Combine both trees
    if(numberSharedElements(clusters(f), clusters(f2)) == 0){
      if(DEBUG >= 1)
        cout <<" --> included" <<endl;
      cliqueTree.msg_pairs.append(candidate_msg_pairs(maxId));
      uint k;
      FOR1D(clusters(f), k){clusters(clusters(f)(k)).setAppend(clusters(f2));}
      FOR1D(clusters(f2), k){clusters(clusters(f2)(k)).setAppend(clusters(f));}
    }
    else {
      // delete if not needed
      delete candidate_msg_pairs(maxId);
      
      if(DEBUG >= 1)
        cout <<" --> discarded" <<endl;
    }
    
    if(DEBUG >= 1){
      cout <<"Elementary clusters [START]" <<endl;
      FOR1D(factors, f){
        cout <<"(" <<f <<") Factor over " <<factors(f)->varIds <<" connected to:" <<endl;
        FOR1D(clusters(f), f2){
          cout <<factors((clusters(f))(f2))->varIds <<"  ";
        }
        cout <<endl;
        // ids only:
        cout<<"In ids: #"<<f<<" -> "<<clusters(f)<<endl;
      }
      cout <<"Elementary clusters [END]" <<endl;
    }
    
    candidate_msg_pairs_contained(maxId) = false;
  }
  
  if(DEBUG >= 1)
    cout <<"THE ALGORITHM [END]" <<endl <<endl;
  
  // delete remaining (unused) msg_pairs
  FOR1D(candidate_msg_pairs, f){
    if(candidate_msg_pairs_contained(f)){
      delete candidate_msg_pairs(f);
    }
  }
  
  // factors in clique tree uebertragen
  cliqueTree.F = factors;
  cliqueTree.setCliqueBeliefs(factors);
}




/**
Jordan, Chap. 17, p.22: "In a junction tree, local consistency implies global consistency."
*/

// bool JunctionTree::checkLocalConsistency(MessagePair& s, bool crashIfInconsistent){
//   uint DEBUG = 1;
//   if(DEBUG > 0)
//     cout <<"Consistency between " <<s.f1->varIds <<" and " <<s.f2->varIds <<" with MessagePair " <<s.varIds;
//   // test marginals over MessagePair variables on both factors
//   Factor f1_marg(s.varIds), f2_marg(s.varIds);
//   tensorMarginal( f1_marg, *(s.f1), s.varIds);
//   tensorMarginal( f2_marg, *(s.f2), s.varIds);
//   if(DEBUG > 0){
//     cout <<endl <<"f1_marg:" <<endl <<f1_marg;
//     cout <<"f2_marg:" <<endl <<f2_marg;
//   }
//   if(f1_marg == f2_marg){
//     if(DEBUG > 0)
//       cout <<" -> consistent" <<endl;
//     // test that messages product is also the same
//     Factor messagesProduct;
//     tensorProduct( messagesProduct, s.m12, s.m21);
//     if(f1_marg == messagesProduct){
//       if(DEBUG > 0)
//         cout <<" Also consistent with m12*m21 = " <<messagesProduct <<"." <<endl;
//       return true;
//     }
//     else {
//       if(DEBUG > 0)
//         cout <<" INCONSISTENT with m12*m21 = " <<messagesProduct <<"." <<endl;
//       HALT("Local inconsistency for "  <<s.f1->varIds <<" and " <<s.f2->varIds);
//       return false;
//     }
//   }
//   else {
//     if(DEBUG > 0)
//       cout <<" -> INCONSISTENT !!!" <<endl;
//     if(crashIfInconsistent){
//       HALT("Local inconsistency for "  <<s.f1->varIds <<" and " <<s.f2->varIds);
//     }
//     return false;
//   }
// }




// uses Marc's eq (12)
// void JunctionTree::jt_passMessage(MessagePair& s, Factor& f_from, FactorGraph& jt){
// 	uint DEBUG = 0;
//     if(DEBUG > 0){
//         cout <<"Pass message [START]" <<endl;
//         cout<<&s<<endl;
//     }
//     
// 	// message m21
// 	if(&f_from == s.f2){
//         if(DEBUG > 0){cout <<f_from.varIds <<" --> " <<s.f1->varIds<<endl;}
// 		if(DEBUG > 2){
// 			cout <<"f_from: " <<endl <<f_from;
//             cout <<"f_to_old: " <<endl <<*s.f1;
//             cout <<"Message 2 --> 1"<<endl;
// 		}
//         
//         // Calculate new msg
//         jt.calcMessage_withDivision(f_from, *s.f1);
// 		// Update target potential (~ incorporate message ~ rescaling)
// 		tensorMultiply(*s.f1, s.m21);
//         
//         if(DEBUG > 1) cout <<"New msg (m21): " <<endl <<s.m21;
//         if(DEBUG > 2) cout <<"f_to: " <<endl <<*s.f1;
// 	}
// 	// message m12
// 	else {
//         if(DEBUG > 0){cout <<f_from.varIds <<" --> " <<s.f2->varIds<<endl;}
// 		if(DEBUG > 2){
//             cout <<"f_from: " <<endl <<f_from <<endl;
//             cout <<"f_to_old: " <<endl <<*s.f2 <<endl;
//             cout <<"Message 1 --> 2"<<endl;
// 		}
// 		
//         // Calculate new msg
//         jt.calcMessage_withDivision(f_from, *s.f2);
// 		// Update target potential (~ incorporate message ~ rescaling)
//         tensorMultiply(*s.f2, s.m12);
// 		
//         if(DEBUG > 1) cout <<"New msg (m12): " <<endl <<s.m12;
//         if(DEBUG > 2) cout <<"f_to: " <<endl <<*s.f2;
// 	}
// 	
// 	if(DEBUG > 0){
// 		cout <<"Pass message [END]" <<endl;
// 	}
// }



/*! Start message collecting from all children-subtrees (i.e., except its parent subtree). !*/
void recursiveCollectEvidence(Factor* node, Factor* parent, FactorGraph& junctionTree){
  uint DEBUG = 0;
  if(DEBUG > 0)
    cout <<"Collect evidence for " <<node->varIds <<" [START]" <<endl;
  uint i;
  FacA neighbors;
  getNeighbors(node, neighbors);
  if(DEBUG > 0){
    cout <<" Neighbors: ";
    printFactorArrayViaVariableSets(neighbors);
  }
  //   cout<<junctionTree<<endl;
  FOR1D(neighbors, i){
    if(neighbors(i) == parent)
      continue;
    recursiveCollectEvidence(neighbors(i), node, junctionTree);
    // update with respect to this direction
    MessagePair* s = getMessagePair(node, neighbors(i));
    // 		cout <<"got MessagePair: " <<std::flush <<s <<endl;
    // message m21
    if(node == s->f1){
      if(DEBUG > 0) cout <<"Pass message " <<s->f2->varIds <<" --> " <<node->varIds <<endl;
      Factor* b_to = junctionTree.B_c(get_list_index_unsigned__orig(junctionTree, s->f1));
      passMessage(*(s->f2), *(s->f1), *b_to, NO_DIV);
    }
    // message 12
    else {
      if(DEBUG > 0) cout <<"Pass message " <<s->f1->varIds <<" --> " <<node->varIds <<endl;
      Factor* b_to = junctionTree.B_c(get_list_index_unsigned__orig(junctionTree, s->f2));
      passMessage(*(s->f1), *(s->f2), *b_to, NO_DIV);
    }
  }
  if(DEBUG > 0)
    cout <<"Collect evidence for " <<node->varIds <<" [END]" <<endl;
}


/*! Starts message distributing on node into the direction of all its children (i.e., except its parent). */
void recursiveDistributeEvidence(Factor* node, Factor* parent, FactorGraph& junctionTree){
  uint DEBUG = 0;
  if(DEBUG > 0)
    cout <<"Distribute evidence for " <<node->varIds <<" [START]" <<endl;
  uint i;
  FacA neighbors;
  getNeighbors(node, neighbors);
  FOR1D(neighbors, i){
    if(neighbors(i) == parent)
      continue;
    // update with respect to this direction
    MessagePair* s = getMessagePair(node, neighbors(i));
    // message m12
    if(node == s->f1){
      // neighbors(i) == s->f2
      if(DEBUG > 0) cout <<"Pass message " <<node->varIds <<" --> " <<s->f2->varIds <<endl;
      Factor* b_to = junctionTree.B_c(get_list_index_unsigned__orig(junctionTree, s->f2));
      passMessage(*(s->f1), *(s->f2), *b_to, NO_DIV);
    }
    // message 21
    else {
      // neighbors(i) == s->f1
      if(DEBUG > 0) cout <<"Pass message " <<node->varIds <<" --> " <<s->f1->varIds <<endl;
      Factor* b_to = junctionTree.B_c(get_list_index_unsigned__orig(junctionTree, s->f1));
      passMessage(*(s->f2), *(s->f1), *b_to, NO_DIV);
    }
    recursiveDistributeEvidence(neighbors(i), node, junctionTree);
  }
  if(DEBUG > 0)
    cout <<"Distribute evidence for " <<node->varIds <<" [END]" <<endl;
}



void JunctionTree::collectAndDistributeInference(FactorGraph& junctionTree){
  // reset junction tree factors to original values
  junctionTree.resetMessages();
  junctionTree.resetCliqueBeliefs();
  
  uint DEBUG = 0;
  if(DEBUG > 0){
    cout <<"---------------------------" <<endl;
    cout <<"Propagating probabilities:" <<endl;
  }
  // determine root	
  Factor* p_root = junctionTree.F(0);
  if(DEBUG > 0){
    cout <<"Root: " <<p_root->varIds <<endl;
  }
  // collect evidence
  if(DEBUG > 0){
    cout <<"\n\n*******************************\n******************************\n**********************\n";
    cout <<"Collecting evidence:" <<endl;
  }
  recursiveCollectEvidence(p_root, 0, junctionTree);
  // distribute evidence
  if(DEBUG > 0){
    cout <<"\n\n*******************************\n******************************\n**********************\n";
    cout <<"Distributing evidence:" <<endl;
  }
  recursiveDistributeEvidence(p_root, 0, junctionTree);
  
  
  //   if(!checkConsistency(junctionTree, true)){
  //     HALT("Propagation of probabilities failed. (Local inconsistencies.)");
  //   }
}


void JunctionTree::constructJunctionTree(FactorGraph& junctionTree, const FacA& factors){
  uint DEBUG = 0;
  if(DEBUG>0){cout<<"constructJunctionTree [START]"<<endl;}
  
  //construct cliques of triangulated tree
  FacA triangulatedCliques;	
  buildTriangulatedCliques(factors, triangulatedCliques);
  
  if(DEBUG > 0){
    cout <<"Input factors:" <<endl;
    write(factors);
    uint f;
    cout <<"Factors built from triangulated graph:" <<endl;
    FOR1D(triangulatedCliques, f){
      cout <<"(" <<f <<") " <<triangulatedCliques(f)->varIds <<endl;
    }
    cout <<endl;
  }
  
  //build junction tree
  buildMaxSpanningTree(triangulatedCliques, junctionTree);
  checkMessagePairConsistency(junctionTree.msg_pairs);
  if(DEBUG>0){cout<<"Message pairs consistent."<<endl;}
  if(DEBUG>0) junctionTree.write(cout);
  junctionTree.resetMessages();
  if(DEBUG > 0){
    cout <<"Junction tree before probability propagation:" <<endl;
    cout <<junctionTree <<endl;
    if(DEBUG > 1)
      junctionTree.writeMessagePairs( cout );
  }
  checkJunctionTreeProperty(junctionTree);
}


void JunctionTree::junctionTreeInference(FactorGraph& junctionTree, const FacA& factors){
  constructJunctionTree(junctionTree,factors);
  collectAndDistributeInference(junctionTree);
  
  uint DEBUG = 0;
  
  if(DEBUG > 0){
    cout <<"Junction tree after probability propagation:" <<endl;
    cout <<junctionTree <<endl;
    if(DEBUG > 1)
      junctionTree.writeMessagePairs( cout );
  }
  if(DEBUG>0){cout<<"constructJunctionTree [END]"<<endl;}
}








//===========================================================================
//
//     Loopy BP methods
//

void LoopyBP::loopy_belief_propagation(FactorGraph& fg, const FacA& factors){
  uint DEBUG = 0;
  constructBipartiteFactorGraph(fg, factors);
  uint MAX_STEPS = 20;
  double change;
  double STOP_TOL = 0.001;
  uint i;
  for(i=0; i<MAX_STEPS; i++){
    if(DEBUG>0){cout<<"+++ Round "<<(i+1)<<" +++"<<endl;}
    change = passAllEdges(fg, PARALLEL);
    if(change < STOP_TOL) break;
    if(DEBUG>0){fg.writeVariableBeliefs();}
    if(DEBUG>1) fg.write(cout);
  }
}


double LoopyBP::passAllEdges(FactorGraph& fg, PassType type){
  if(type == PARALLEL){
    return passAllEdges_parallel(fg);
  }
  else
    NIY;
  return 0;
}


double LoopyBP::passAllEdges_parallel(FactorGraph& fg){
  uint DEBUG =0;
  // (1) cliques --> vars
  uint i;
  if(DEBUG>0){cout<<"***** Propagating forth (factor -> variable) *****"<<endl;}
  FOR1D(fg.F, i){
    shoutMessages(*fg.F(i), NO_DIV);
  }
  if(DEBUG>0){cout<<"***** Propagating back (variable -> factor) *****"<<endl;}
  // (2) vars --> cliques
  FOR1D(fg.F_v, i){
    shoutMessages(*fg.F_v(i), NO_DIV);
  }
  double maxChange = fg.computeBeliefs();
  if(DEBUG>0){cout<<"Belief change:  maxChange="<<maxChange<<endl;}
  return maxChange;
}

void LoopyBP::shoutMessages(Factor& f, MsgCalc calcMsgType){
  uint i;
  FOR1D(f.msg_pairs, i){
    Factor* f_to;
    if(f.msg_pairs(i)->f1 == &f)
      f_to = f.msg_pairs(i)->f2;
    else
      f_to = f.msg_pairs(i)->f1;
    if(calcMsgType == WITH_DIV)
      computeMessage_withDiv(f, *f_to);
    else
      computeMessage_noDiv(f, *f_to);
  }
}



void check_exactlyOneConditional(VarA& vars, FacA& facs){
  uint i, k;
  CHECK(vars.N==facs.N, "#vars != #facs");
  FOR1D(vars, i){
    FOR1D(facs, k){
      if(facs(k)->varIds(0) == vars(i)->id) break;
    }
    CHECK(k<facs.N, "No factor for variable "<<vars(i)->id);
  }
}

void check_atLeastOneConditional(VarA& vars, FacA& facs){
  uint i, k;
  FOR1D(vars, i){
    FOR1D(facs, k){
      if(facs(k)->varIds(0) == vars(i)->id) break;
    }
    CHECK(k<facs.N, "No factor for variable "<<vars(i)->id);
  }
}






// =======================================================================
//
//  Exact Tree Inference (by MT for NP)
//

/* message orders allows us to compute a series of messages
   in a specified order and direction. A message order is stored
   as a list of messages and a directionality with each of them. */
   

std::ostream& operator<<(std::ostream& os,const MessageOrderEntry& moe){
  os <<"moe.m=";
  moe.m->writeIds(os);
  return os <<" moe.fwd=" <<moe.fwd <<endl;
}

//! compute a series of messages in a given order, optionally inversely (also inverting each message's directionality)
void recomputeBatchOfMessages(MessageOrder& order,bool invert_order){
  uint i,N=order.N;
  if(!invert_order) for(i=0;i<N;i++){
    if(order(i).fwd) recomputeMessage_12(*order(i).m);
    else             recomputeMessage_21(*order(i).m);
  }else for(i=N;i--;){
    if(order(i).fwd) recomputeMessage_21(*order(i).m);
    else             recomputeMessage_12(*order(i).m);
  }
}

//! for inference in a tree, we first construct an ordering descending from the root
void constructTreeMessageOrder(MessageOrder& order,const Factor *root){
  uint i,j;
  Factor *f;
  MessagePair *m;

  //--- construct a tree oder
  //first ad all neighbor links of root to the order
  for(j=0;j<root->msg_pairs.N;j++){
    m=root->msg_pairs(j);
    if(m->f1==root) order.append(MessageOrderEntry(m,true));
    else            order.append(MessageOrderEntry(m,false));
  }
  //step through the existing links and append sublinks dynamically
  for(i=0;i<order.N;i++){
    if(order(i).fwd) f=order(i).m->f2; else f=order(i).m->f1; //f is the sub-factor of order(i).m
    for(j=0;j<f->msg_pairs.N;j++){
      m=f->msg_pairs(j);                                     //m is one sub-message of f
      if(m==order(i).m) continue;                             //discard it when it is going up
      if(m->f1==f) order.append(MessageOrderEntry(m,true));
      else         order.append(MessageOrderEntry(m,false));
    }
  }
}

//! inference on a tree is then trivial: compute a descending-from-root order,
//! pass all messages backward (collecting towards the root)
//! pass all messages forward  (distributing from the root)
void treeInference(const Factor *root,bool check_consitency){
  checkSpaceRegistry(0);
  MessageOrder order;

  constructTreeMessageOrder(order,root);

  //cout <<order <<endl;

  recomputeBatchOfMessages(order,true);
  recomputeBatchOfMessages(order,false);

  checkSpaceRegistry(0);

  if(check_consitency){
    for(uint i=0;i<order.N;i++) checkConsistency(*order(i).m);
  }
  MT_MSG("everything consistent :-)");
}


// =======================================================================
//
//  inference for mixture length DBNs
//

void inferMixLengthUnstructured(
    arr& alpha, arr& beta, arr& PT, double& PR, double& ET,
    const arr& S, const arr& R, const arr& P, double gamma, uint Tmax){
  DEBUG_INFER(1,cout <<MT_HERE <<endl);
  arr a,b;
  double gt,gSum;
  uint t;
  CHECK(S.nd==1 && R.nd==1 && P.nd==2 && S.N==R.N && S.N==P.d0 && S.N==P.d1,"");
  CHECK(gamma>0. && gamma<=1.,"");
  a = S;
  b = R;
  alpha = a;
  beta  = b;
  PT.resize(2*Tmax+1);
  PT(0) = scalarProduct(a,b);
  gt = gamma;
  for(t=1;t<=Tmax;t++){
    a = P*a;
    PT(2*t-1) = scalarProduct(a,b);
    b = b*P;
    PT(2*t) = scalarProduct(a,b);
    alpha += gt*a;
    gt *= gamma;
    beta += gt*b;
  }
  PR = 0.;
  ET = 0.;
  gt = 1.;
  gSum=0.;
  for(t=0;t<=2*Tmax;t++){
    PT(t) *= gt;
    ET += t*PT(t);
    PR += PT(t);
    gSum += gt;
    gt *= gamma;
  }
  PT /= PR;
  ET /= PR;
  alpha *= (1.-gamma);
  PR    *= (1.-gamma);
}

void inferMixLengthStructured(
    Factor& alpha, Factor& beta, arr& PT, double& PR, double& ET,
    const uintA& headVars, const uintA& tailVars,
    const FacA& S, const FacA& R, const FacA& P, double gamma, uint Tmax){
  DEBUG_INFER(1,cout <<MT_HERE <<endl);
  Factor a,b;
  double gt,gSum;
  uint t;
  DEBUG_INFER(2,cout <<"  headVars=" <<headVars <<"  tailVars=" <<tailVars <<endl);
  CHECK(headVars.N==tailVars.N,""); //actually should also check that their dims are equal...
  CHECK(gamma>0. && gamma<=1.,"");
  //get initial a and b:
  eliminationAlgorithm(a,S,headVars);
  //FacA tmp = P;  tmp.append(R);
  eliminationAlgorithm(b,R,headVars);  b.varIds=tailVars;
  //above, we reassociate the factor to the tail variables although it
  //was initially defined over the head variables
  //we do this same ``trick'' (or hack) also below
  alpha = a;
  beta  = b;
  PT.resize(2*Tmax+1);
  PT(0) = scalarProduct(a.P,b.P)*::exp(a.logP+b.logP);
  gt = gamma;
  FacA fwdList = P;  fwdList.append(&a);
  FacA bwdList;      bwdList.append(&b); bwdList.append(P);
  for(t=1;t<=Tmax;t++){
    eliminationAlgorithm(a,fwdList,tailVars);  a.varIds=headVars;
    PT(2*t-1) = scalarProduct(a.P,b.P)*::exp(a.logP+b.logP);
    eliminationAlgorithm(b,bwdList,headVars);  b.varIds=tailVars;
    PT(2*t)   = scalarProduct(a.P,b.P)*::exp(a.logP+b.logP);
    tensorWeightedAdd(alpha,gt,a);
    gt *= gamma;
    tensorWeightedAdd(beta ,gt,b);
  }
  PR = 0.;
  ET = 0.;
  gt = 1.;
  gSum=0.;
  for(t=0;t<=2*Tmax;t++){
    PT(t) *= gt;
    ET += t*PT(t);
    PR += PT(t);
    gSum += gt;
    gt *= gamma;
  }
  PT /= PR;
  ET /= PR;
  alpha.logP += ::log(1.-gamma);
  PR    *= (1.-gamma);
}
    
