#include "infer_tree.h"
#include "infer.h"

std::ostream& operator<<(std::ostream& os,const TreeNode& t){
  return os <<"par=" <<t.parent <<" dim=" <<t.dim <<" P=" <<t.P <<endl;
}

void randomTree(Tree& tree, uint N, uint K, uint roots){
  tree.resize(N);
  uint n;
  for(n=0;n<N;n++){
    tree(n).dim=K;                 //K-dim variable
    if(n<roots) tree(n).parent=-1;
    else        tree(n).parent=rnd(n);  //random parent
    if(n<roots) tree(n).P.resize(K);    //K-times-K joint factor (this,parent)
    else        tree(n).P.resize(K,K);
    MT_MSG("fully random roots!");
    /*if(n<roots) { // deterministic roots for testing purposes
      tree(n).P(0) = 0.0;
      tree(n).P(1) = 1.0;
    } else*/
    rndUni(tree(n).P,.1,1.,false);
  }
}


/* conversion: every tree node corresponds to
- a variable
- a pairwise factor connecting it to its parent
- and a Message connecting its factor to the parent's factor

everything is ordered precisely */

void tree2FactorGraph(FactorGraph& fg,const MT::Array<TreeNode>& tree){
  uint i,N=tree.N;
  fg.V.resize(N);
  fg.F.resize(N);
  for(i=0;i<N;i++){    //variables
    fg.V(i) = new Variable(tree(i).dim, STRING("tree_node_"<<std::setfill('0')<<std::setw(3)<<i));
  }
  for(i=0;i<N;i++){    //factors
    if(tree(i).parent<0) fg.F(i) = new Factor(TUP(fg.V(i)->id));
    else                 fg.F(i) = new Factor(TUP(fg.V(i)->id,fg.V(tree(i).parent)->id));
    fg.F(i)->set(tree(i).P);
  }
  for(i=0;i<N;i++){    //messages
    if(tree(i).parent>=0) fg.msg_pairs.append(new MessagePair(fg.F(i),fg.F(tree(i).parent)));
  }
  fg.resetMessages();
  fg.resetCliqueBeliefs();
}

// void getPosteriors(MT::Array<arr>& posteriors, FactorGraph& model) {
//   uint i;
//   FOR1D(model.V, i) {
//     uintA var_ids;
//     var_ids.append(model.V(i)->id);
//     Factor marginal;
//     getMarginal(marginal, var_ids, model);
//     posteriors.append(marginal.P);
//   }
// }

void treeInference(MT::Array<arr>& posteriors, const Tree& tree){
  FactorGraph model;
  tree2FactorGraph(model,tree);
  treeInference(model.F(0),true);
  getVariableBeliefs(posteriors, model.V);
}

void treeInference(MT::Array<arr>& posteriors,const Tree& forest,uintA& roots){
  FactorGraph model;
  tree2FactorGraph(model,forest);
  for(uint i=0;i<roots.N;i++){
    treeInference(model.F(roots(i)),true);
  }
//   model.write(cout, true, true);
//   model.writeMessagePairs(cout);
  getVariableBeliefs(posteriors, model.V);
}


void write(Tree& tree) {
  uint i;
  cout<<"tree: ";
  FOR1D(tree, i) {
    cout<<tree(i).parent<<" ";
  }
  cout<<endl;
}
