#include "mdp.h"
#include <vector>
#include <map>

byteA global_maze;

void writeMDP(const char *filename, MDP &mdp,bool binary){
  ofstream os;
  if(binary) os.open(filename,std::ios::binary);
  else os.open(filename);
  writeArr(mdp.Pxax,"Pxax",os);
  writeArr(mdp.Pyxa,"Pyxa",os);
  writeArr(mdp.Px,"Px",os);
  writeArr(mdp.Rax,"Rax",os);
  writeArr(arr(&mdp.gamma,1),"gamma",os);
}

void readMDP(const char *filename, MDP &mdp,bool binary){
  ifstream is;
  if(binary) is.open(filename,std::ios::binary);
  else is.open(filename);
  readArr(mdp.Pxax,"Pxax",is);
  readArr(mdp.Pyxa,"Pyxa",is);
  readArr(mdp.Px,"Px",is);
  readArr(mdp.Rax,"Rax",is);
  readArr(arr(&mdp.gamma,1)(),"gamma",is);
}

void convertTonyToMDP(const char* POMDP_file){
  MDP mdp;
  readTonyMDP(POMDP_file, mdp.Pxax, mdp.Pyxa, mdp.Px, mdp.Rax, mdp.gamma);
  writeMDP(STRING(POMDP_file<<".arr"),mdp);
}

void reportMDP(const arr& Pxax,const arr& Pyxa){
  uint xx,y,x,a;
  //
  uintA xm(Pxax.d0);
  for(a=0;a<Pxax.d1;a++){
    xm.reshape(Pxax.d0); xm.setZero();
    for(x=0;x<Pxax.d0;x++) for(xx=0;xx<Pxax.d2;xx++)
      if(Pxax(xx,a,x)>Pxax(xm(x),a,x)) xm(x)=xx;
      xm.reshape(global_maze.d0,global_maze.d1);
      cout <<"\nmax transitions for a=" <<a <<"\n" <<xm;
  }
  
  //
  uintA ym(Pyxa.d1); ym.setZero();
  for(y=0;y<Pyxa.d0;y++)
    for(x=0;x<Pyxa.d1;x++)
      for(a=0;a<Pyxa.d2;a++) if(Pyxa(y,x,a)>Pyxa(ym(x),x,a)) ym(x)=y;
      ym.reshape(global_maze.d0,global_maze.d1);
      cout <<"\nmax observations =\n" <<ym;
}

void readImageMaze(const char* filename,MDP &mdp){
  byteA img;
  read_ppm(img,filename);
  byteA maze(img.d0,img.d1);
  uint x,y,dx=maze.d1,dy=maze.d0;
  for(x=0;x<dx;x++) for(y=0;y<dy;y++){
    if(img(y,x,0)==255 && img(y,x,1)==255 && img(y,x,2)==255) maze(y,x)=0; //white -> free
    else if(img(y,x,0)==0   && img(y,x,1)==0   && img(y,x,2)==0  ) maze(y,x)=1; //black -> wall
    else if(img(y,x,0)==255 && img(y,x,1)==0   && img(y,x,2)==0  ) maze(y,x)=2; //red   -> start
    else if(img(y,x,0)==0   && img(y,x,1)==255 && img(y,x,2)==0  ) maze(y,x)=3; //green -> goal
    else if(img(y,x,0)==0   && img(y,x,1)==0   && img(y,x,2)==255) maze(y,x)=4; //blue  -> avoid
    else HALT("maze image file (ppm) needs to be black/white/red/green/blue");
  }
  mazeToMDP(maze,mdp);
}

void mazeToMDP(const byteA& maze,MDP& mdp){
  mazeToP(maze,mdp.Px,mdp.Pxax,mdp.Pyxa,mdp.Rax);
}

void mazeToP(const byteA& maze,arr& Px,arr& Pxax,arr& Pyxa,arr& Rax,bool trapWalls){
  CHECK(maze.nd==2,"");
  global_maze = maze;
  uint nx=maze.N,na=5,ny=16,dx=maze.d1,dy=maze.d0;
  uint x,a,i,j,k;
  
  Pxax.resize(nx,na,nx);  Pxax.setZero();
  Pyxa.resize(ny,nx,na);  Pyxa.setZero();
  Px  .resize(nx);        Px  .setZero();
  Rax .resize(na,nx);     Rax .setZero();
  
  /*
  0: free space
  1: wall
  2: start state (can be multiple)
  3: positive reward +1
  4: negative reward -1
  */
  
  for(x=0;x<nx;x++){
    i=x/dx; j=x%dx;
    if(maze(i,j)==1){
      Pxax(x,0,x)=Pxax(x,1,x)=Pxax(x,2,x)=Pxax(x,3,x)=Pxax(x,4,x)=1.;
      for(a=0;a<na;a++) Pyxa(0,x,a)=1.;
    }else{
      k=0; //observation indicator
      if(i>0    && (trapWalls || maze(i-1,j)!=1)){ Pxax(x-dx,0,x)=1.; }else{ k|=1; Pxax(x,0,x)=1.; }//up
      if(j<dx-1 && (trapWalls || maze(i,j+1)!=1)){ Pxax(x+1 ,1,x)=1.; }else{ k|=2; Pxax(x,1,x)=1.; }//right
      if(i<dy-1 && (trapWalls || maze(i+1,j)!=1)){ Pxax(x+dx,2,x)=1.; }else{ k|=4; Pxax(x,2,x)=1.; }//down
      if(j>0    && (trapWalls || maze(i,j-1)!=1)){ Pxax(x-1 ,3,x)=1.; }else{ k|=8; Pxax(x,3,x)=1.; }//left
      Pxax(x,4,x) = 1.; //stay
      for(a=0;a<na;a++) Pyxa(k,x,a)=1.; //observe `k'
    }
    if(maze(i,j)==2){
      Px(x)=1.;
    }
    if(maze(i,j)==3){
      for(a=0;a<na;a++) Rax(a,x)=1.;
    }
    if(maze(i,j)==4){
      for(a=0;a<na;a++) Rax(a,x)=-1.;
    }
  }
  
  normalizeDist(Px);
  
  checkNormalization(Pxax);
  checkNormalization(Pyxa);
  checkNormalization(Px);
}

void addActionNoise(arr &Pxax,double eps){
  uint nx=Pxax.d0,na=Pxax.d1;
  uint i,j,a;
  arr Pxx;
  tensorMarginal(Pxx,Pxax,TUP(0,2));
  Pxx /= (double)na;
  checkNormalization(Pxx);
  for(i=0;i<nx;i++) for(j=0;j<nx;j++) for(a=0;a<na;a++){
    Pxax(i,a,j) = (1.-eps) * Pxax(i,a,j) + eps * Pxx(i,j);
  }
  checkNormalization(Pxax);
}

void tunnel(arr& Pxax,int from,int to){
  uint nx=Pxax.d0,na=Pxax.d1;
  if(from<0) from+=nx;
  if(to  <0) to  +=nx;
  uint x,a;
  for(a=0;a<na;a++){
    for(x=0;x<nx;x++) Pxax(x,a,from)=0;
    Pxax(to,a,from)=1;
  }
  checkNormalization(Pxax);
}

void tunnelToPx(arr& Pxax,int from,const arr& Px){
  uint nx=Pxax.d0,na=Pxax.d1;
  if(from<0) from+=nx;
  uint x,a;
  for(a=0;a<na;a++){
    for(x=0;x<nx;x++) Pxax(x,a,from)=Px(x);
  }
  checkNormalization(Pxax);
}

void tunnelRaxToPx(arr& Pxax,const arr& Rax,const arr& Px){
  uint nx=Pxax.d0,na=Pxax.d1;
  uint x,y,a;
  for(x=0;x<nx;x++) for(a=0;a<na;a++){
    if(Rax(a,x)){
      for(y=0;y<nx;y++) Pxax(y,a,x)=Px(y);
    }
  }
  checkNormalization(Pxax);
}

void tunnelRaxTo(arr& Pxax,const arr& Rax,int to){
  uint nx=Pxax.d0,na=Pxax.d1;
  if(to  <0) to  +=nx;
  uint x,y,a;
  for(x=0;x<nx;x++) for(a=0;a<na;a++){
    if(Rax(a,x)){
      for(y=0;y<nx;y++) Pxax(y,a,x)=0.;
      Pxax(to,a,x)=1;
    }
  }
  checkNormalization(Pxax);
}

void neutralSplit(arr& P0y0,arr& Pa0,uint i,double eps){
  uint n0=P0y0.d0,ny=P0y0.d1,na=Pa0.d0;
  uint a,y,j;
  
  //make them bigger by inserting zeros
  P0y0.reshape(n0,ny*n0);
  P0y0.insRows(n0,1);
  P0y0.reshape((n0+1)*ny,n0);
  P0y0.insColumns(n0,1);
  Pa0.insColumns(n0,1);
  P0y0.reshape(n0+1,ny,n0+1);
  Pa0.reshape(na,n0+1);
  
  //copy the output of the ith node indentically
  for(a=0;a<na;a++) Pa0(a,n0) = Pa0(a,i);
  for(j=0;j<n0;j++) for(y=0;y<ny;y++) P0y0(j,y,n0) = P0y0(j,y,i);
  checkNormalization(Pa0);
  checkNormalization(P0y0);
  
  //split inputs to new node between i and new node
  double p,s;
  for(j=0;j<n0;j++) for(y=0;y<ny;y++){
    p=P0y0(i,y,j);
    s=.5+rnd.uni(-eps,eps);
    P0y0(i,y,j)  = s*p;
    P0y0(n0,y,j) = (1.-s)*p;
  }
  checkNormalization(P0y0);
}

void generateStandardProblem(uint problem,MDP& mdp){
  generateStandardProblem(problem,mdp.Px,mdp.Pxax,mdp.Pyxa,mdp.Rax);
}

void generateStandardProblem(uint problem,arr& Pxax,arr& Pyxa,arr& Px,arr& Rax){
  byteA maze;
  switch(problem){
  case tinyMaze:
  maze <<"[\
         2 1;\
         0 3]";
  maze-=(byte)'0';
  
  mazeToP(maze,Pxax,Pyxa,Px,Rax);
  tunnel(Pxax,-1,1);
  
  //addActionNoise(Pxax,.1);
  break;
  case simpleMaze:
  maze <<"[\
         2 1 0 0 0 1 0 0 0;\
         0 1 0 1 0 1 0 1 0;\
         0 1 0 1 0 1 0 1 0;\
         0 1 0 1 0 1 0 1 0;\
         0 0 0 1 0 0 0 1 3]";
  
         /*maze <<"[\
         1 1 1 1 1 1 1;\
         1 3 0 0 0 0 1;\
         1 1 1 1 1 0 1;\
         1 2 0 0 0 0 1;\
  1 1 1 1 1 1 1]";*/
  maze-=(byte)'0';
  
  mazeToP(maze,Pxax,Pyxa,Px,Rax);
  tunnel(Pxax,-1,-2);
  
  //addActionNoise(Pxax,.1);
  break;
  case heavenAndHell:
  maze <<"[\
         3 0 4 1 4 0 3 ;\
         1 2 1 1 1 2 1 ;\
         0 0 1 1 1 0 0 ]";
  maze-=(byte)'0';
  
  mazeToP(maze,Pxax,Pyxa,Px,Rax);
  tunnel(Pxax,0,7);
  tunnel(Pxax,6,13);
  
  tunnel(Pxax,2,9);
  tunnel(Pxax,4,11);
  
  tunnelToPx(Pxax,0,Px);
  tunnelToPx(Pxax,6,Px);
  tunnelToPx(Pxax,2,Px);
  tunnelToPx(Pxax,4,Px);
  
  //addActionNoise(Pxax,.1);
  break;
  default:
    HALT("don't know problem "<<problem);
  }
}

void createNeighorList(MDP& mdp){
  uint x,y,a,A=mdp.Pxax.d1,X=mdp.Px.d0;
  mdp.neighbors.resize(X);
  for(y=0;y<X;y++) for(x=0;x<=y;x++) for(a=0;a<A;a++) {
    if(mdp.Pxax(y,a,x)!=0. || mdp.Pxax(x,a,y)!=0){
      mdp.neighbors(x).append(y);
      if(x!=y) mdp.neighbors(y).append(x);
      a=A;
    }
  }
}

//===========================================================================
//
// collapsing controllers into a 1-level controller
//

void collapse2levelFSC(FSC_lev1& fsc1, const FSC_lev2& fsc2){
  
  uint d0=fsc2.P0.d0, d1=fsc2.P1.d0, da=fsc2.Pa0.d0, dy=fsc2.P1y01.d1;
  uint i,j,k;
  
  fsc1.P0.resize(d0,d1);
  tensorEquation(fsc1.P0, fsc2.P0, TUP(0), fsc2.P1, TUP(1));
  fsc1.P0.reshape(d0*d1);
  
  fsc1.Pa0.resize(da,d0,d1);
  for(i=0;i<da;i++) for(j=0;j<d0;j++) for(k=0;k<d1;k++)
    fsc1.Pa0(i,j,k) = fsc2.Pa0(i,j);
  fsc1.Pa0.reshape(da,d0*d1);
  
  fsc1.P0y0.resize(TUP(d0,d1,dy,d0,d1));
  tensorEquation(fsc1.P0y0, fsc2.P01y0, TUP(0,1,2,3), fsc2.P1y01, TUP(1,2,3,4));
  fsc1.P0y0.reshape(d0*d1,dy,d0*d1);
  
  tensorCheckCondNormalization(fsc1.P0,1);
  tensorCheckCondNormalization(fsc1.Pa0,1);
  tensorCheckCondNormalization(fsc1.P0y0,1);
};

//===========================================================================
//
// standard initializations
//

void oneNodeOneAction(arr& Pa0,uint da,uint d0,double uni,double noise,double det){
  Pa0.resize(da,d0);
  Pa0 = uni;
  rndUni(Pa0,0.,noise,true);
  for(uint n=1;n<d0;n++) Pa0((n-1)%da,n)+=det;
  tensorCondNormalize(Pa0,1);
  tensorCheckCondNormalization(Pa0,1);
}

void zeroNodeStart(arr& P0,uint d0,double uni,double noise,double zero){
  P0.resize(d0);
  P0 = uni;
  rndUni(P0,.0,noise,true);
  P0(0)+=zero;
  tensorCondNormalize(P0,1);
  tensorCheckCondNormalization(P0,1);
} 

void generalRandTransitions(arr& P,double uni,double noise){
  P = uni;
  rndUni(P,.0,noise,true);
  tensorCondNormalize(P,1);
  tensorCheckCondNormalization(P,1);
}

void generalNodeTransitions(arr& P0y0,uint d0,uint dy,double uni,double noise,double stay){
  P0y0.resize(d0,dy,d0);
  uint i,j;
  P0y0 = uni;
  rndUni(P0y0,.0,noise,true);
  for(i=0;i<d0;i++) for(j=0;j<dy;j++) P0y0(i,j,i)+=stay;
  tensorCondNormalize(P0y0,1);
  tensorCheckCondNormalization(P0y0,1);
}

void generalNode0Transition(arr& P01y0,
                            uint d0,uint d1,uint dy,
                            double uni,double noise,double stay){
  P01y0.resize(TUP(d0,d1,dy,d0));
  uint i,j,k;
  P01y0 = uni;
  rndUni(P01y0,.0,noise,true);
  for(i=0;i<d0;i++) for(j=0;j<d1;j++) for(k=0;k<dy;k++) P01y0(TUP(i,j,k,i))+=stay;
  tensorCondNormalize(P01y0,1);
  tensorCheckCondNormalization(P01y0,1);
}

void generalNode1Transition(arr& P1y01,
                            uint d1,uint dy,uint d0,
                            double uni,double noise,double stay){
  P1y01.resize(TUP(d1,dy,d0,d1));
  uint i,j,k;
  P1y01 = uni;
  rndUni(P1y01,.0,noise,true);
  for(i=0;i<d1;i++) for(j=0;j<dy;j++) for(k=0;k<d0;k++) P1y01(TUP(i,j,k,i))+=stay;
  tensorCondNormalize(P1y01,1);
  tensorCheckCondNormalization(P1y01,1);
}


//===========================================================================
//
// strict hierarchies
//

void initHierarchical(uint exits,FSC_lev2& fsc){
  uint i;
  uint d0=fsc.P01y0.d0, d1=fsc.P1y01.d0, dy=fsc.P1y01.d1;
  cout <<"building hierarchical controller: d0=" <<d0 <<" d1=" <<d1 <<" dy=" <<dy
    << " -- assuming the last "<< exits<< " level-0-nodes to be exits" <<endl;
  
  fsc.hierarchical=true;
  
  fsc.Pe0.resize(2,d0);         //P(e|n0)  is fixed!
  fsc.PE_01.resize(d0,d1);      //P(n0'|n1'; exit=true)
  fsc.P_E_0y0.resize(d0,dy,d0); //P(n0'|y',n0; exit=false)
  fsc.PE_1y1.resize(d1,dy,d1);  //P(n1'|y',n1; exit=true)
  fsc.P_E_11.resize(d1,d1);     //P(n1'|n1; exit=false) is fixed!
  
  fsc.Pe0=1.;    //no constraints on which n0 trigger an exit state?
  for(i=0;i<d0;i++){
    if(i>=d0-exits){ fsc.Pe0(1,i)=1.; fsc.Pe0(0,i)=0.; }
    else           { fsc.Pe0(1,i)=0.; fsc.Pe0(0,i)=1.; }
  }
  tensorCheckCondNormalization(fsc.Pe0,1);
  
  fsc.P_E_11.setId();
  
  generalRandTransitions(fsc.PE_01,1.,1.);
  generalRandTransitions(fsc.P_E_0y0,1.,1.);
  generalRandTransitions(fsc.PE_1y1,1.,1.);
  
  hierarchicalToFactors(fsc);
}

void hierarchicalToFactors(FSC_lev2& fsc){
  CHECK(fsc.hierarchical,"");
  uint i,j,k,l;
  uint d0=fsc.P01y0.d0, d1=fsc.P1y01.d0, dy=fsc.P1y01.d1;
  
  arr P01ye0;
  P01ye0.resize(TUP(d0,d1,dy,2,d0));
  for(i=0;i<d0;i++) for(j=0;j<d1;j++) for(k=0;k<dy;k++) for(l=0;l<d0;l++){
    P01ye0(TUP(i,j,k,0,l)) = fsc.P_E_0y0(i,k,l);
    P01ye0(TUP(i,j,k,1,l)) = fsc.PE_01(i,j);
  }
  tensorCheckCondNormalization(P01ye0,1);
  
  arr P1ye1;
  P1ye1.resize(TUP(d1,dy,2,d1));
  for(i=0;i<d1;i++) for(j=0;j<dy;j++) for(k=0;k<d1;k++){
    P1ye1(TUP(i,j,0,k)) = fsc.P_E_11(i,k);
    P1ye1(TUP(i,j,1,k)) = fsc.PE_1y1(i,j,k);
  }
  tensorCheckCondNormalization(P1ye1,1);
  
  tensorEquation(fsc.P1y01, P1ye1, TUP(0,1,4,3), fsc.Pe0, TUP(4,2), 1);
  tensorEquation(fsc.P01y0, P01ye0, TUP(0,1,2,4,3), fsc.Pe0, TUP(4,3), 1);
  tensorCheckCondNormalization(fsc.P1y01,1);
  tensorCheckCondNormalization(fsc.P01y0,1);
}


//===========================================================================
//===========================================================================


void OutputAoDot(arr &Pa0, uint numObs, const char* problem, std::vector<std::string> &a, 
                 std::vector<std::string> &o, 
                 std::map<int, std::string> &a0) {
  uint i,j;
  cout << problem << endl;
  if(!strcmp(problem,"chainOfChains3.POMDP.arr")) {
    a.push_back("doNothing");a.push_back("wetHands"); a.push_back("dryHands");
    a.push_back("soapyHands");a.push_back("rinseHands");a.push_back("waterOn");
    a.push_back("waterOff");
    //  ta: taps, w: water, so: soap, to: towel, si: sink, a: away 
    o.push_back("ota");o.push_back("fta");o.push_back("ow"); o.push_back("fw");
    o.push_back("oso"); o.push_back("fso"); o.push_back("oto"); o.push_back("fto");
    o.push_back("osi"); o.push_back("fsi"); o.push_back("oa"); o.push_back("fa");
  }
  else if(!strcmp(problem,"chainOfChains3.POMDP.arr")) {
    a.push_back("A");a.push_back("B");a.push_back("C");a.push_back("D");
    o.push_back("1"); 
  }
  else if(!strcmp(problem,"cheese-taxi-fully-observable.POMDP.arr")) {
    a.push_back("N");a.push_back("S");a.push_back("E");a.push_back("W");
    a.push_back("query");a.push_back("pickup");a.push_back("putdown");
    o.push_back("s0"); o.push_back("s1"); o.push_back("s2"); o.push_back("s3");
    o.push_back("s4"); o.push_back("s5"); o.push_back("s6"); o.push_back("s7");
    o.push_back("s8"); o.push_back("s9"); o.push_back("s10"); o.push_back("s0d1");
    o.push_back("s1d1"); o.push_back("s2d1"); o.push_back("s3d1"); o.push_back("s4d1");
    o.push_back("s5d1"); o.push_back("s6d1"); o.push_back("s7d1"); o.push_back("s8d1");
    o.push_back("s9d1"); o.push_back("s10d1"); o.push_back("s0d2"); o.push_back("s1d2");
    o.push_back("s2d2"); o.push_back("s3d2"); o.push_back("s4d2"); o.push_back("s5d2");
    o.push_back("s6d2"); o.push_back("s7d2"); o.push_back("s8d2"); o.push_back("s9d2");
    o.push_back("s10d2"); o.push_back("d0"); o.push_back("d1"); o.push_back("d2");
  }
  else { // Simply put placeholders
    for(i=0;i<Pa0.d0;++i) { 
      std::stringstream s;
      s << i;
      a.push_back(s.str());
    }
    for(i=0;i<numObs;++i) { 
      std::stringstream s;
      s << i;
      o.push_back(s.str());
    }
  }
  
  // Map Nodes to Actions 
  for(j=0;j<Pa0.d1;++j) { // j is the node
    for(i=0;i<Pa0.d0;++i) { // i is the action
      double value= Pa0(i,j);
      if(value > 0.01) {
        if(value > 0.99) value = 1.0;
        //std::stringstream ss; ss<<value; std::string str; ss>>str;
        std::stringstream ss; ss<< j; std::string str; ss>>str;
        //a0[j] = a[i]+"_"+str+"_"+a0[j];
        a0[j] = str+a[i]+"_"+a0[j];
  } } }
}

// Outputs a .dot file. 
// To convert it to a postscript: $ dot -Tps filename.dot -o filename.ps
void OutputDotH(const char* filename, const char* problem, arr &Pa0, arr &Pe0, arr &PE_01, arr &P_E_0y0, arr &PE_1y1, 
                arr &P_E_11) { 
  uint i,j;
  std::vector<std::string> a, o;
  std::map<int, std::string> a0;
  OutputAoDot(Pa0, P_E_0y0.d1, problem, a, o, a0);
  
  cout << "Outputing graph: " << filename << endl;
  // Output graph
  ofstream out(filename);
  out << "digraph " << "ML_Controller_" << Pa0.d1 << "_" << PE_1y1.d0 << " {\n"
    << "\trankdir=TB;\n";
  
  // Level-0
  out << "\tsubgraph cluster_0 {\n" 
    << "\t\tstyle=filled;\n"
    << "\t\tcolor=lightgrey;\n"
    << "\t\tnode [style=filled,color=white];\n\t\t{rank = same;"; 
  for(i=0;i<Pa0.d1;++i) {
    // check if it's an exit state
    // if so, output it as a diamond....
    if(Pe0(1,i)<0.01)
      out << "N0_" << a0[i] << "; ";
  }
  out << "};\n"
    << "\t\tlabel=\"Level 0\"\n";
  // Level-Exits
  out << "\tsubgraph cluster_2 {\n" 
    << "\t\tstyle=filled;\n"
    << "\t\tcolor=lightgrey;\n"
    << "\t\tnode [style=filled,color=white,shape=diamond];"; 
  for(i=0;i<Pa0.d1;++i) { if(Pe0(1,i)>0.99) out << "E_" << a0[i] << " ";}
  out << ";\n"
    << "\t\tlabel=\"Exit\"\n"
    << "\t}\n"
    << "\n\t}\n";
  
  // Level-1
  out << "\tsubgraph cluster_1 {\n" 
    << "\t\tstyle=filled;\n"
    << "\t\tcolor=lightgrey;\n"
    // This is correct but the rank=same makes dot crash once in a while, this
    // is a dot bug.
    << "\t\tnode [style=filled,color=white];\n\t\t{rank = same;"; for(i=0;i<PE_01.d1;++i) {out << "N1_" << i << "; ";}
  //<< "\t\tnode [style=filled,color=white]; "; for(i=0;i<PE_01.d1;++i) {out << "N1_" << i << "; ";}
  out << "};\n"
    << "\t\tlabel=\"Level 1\"\n"
    << "\t}\n";
  
  
  // P_E_0y0; E is false
  for(i=0;i<P_E_0y0.d0;++i) // incoming node is i
    for(uint k=0;k<P_E_0y0.d2;++k) { // outgoing node is k
      bool obs = false;
      if(!(Pe0(1,k)>0.99))
        for(uint j=0;j<P_E_0y0.d1;++j) { // outgoing obs. is j
          // Merge observations 
          if(P_E_0y0(i,j,k) > 0.01) { 
            if(!obs) { 
              out << "\tN0_" << a0[k] << ((Pe0(1,i)>0.99)? "-> E_":"-> N0_") << a0[i] << " [label=\""; 
              obs=true;
            }
            if (P_E_0y0(i,j,k) < 0.99) {
              out << "" << o[j] << ":" << std::setprecision(2) 
                << P_E_0y0(i,j,k) << ",";
            }
          }
        }
        if(obs) out << "\",arrowsize=0.3,labelfontsize=9.0,penwidth=0.5];\n";
    }
    
    // PE_1y1; E is true
    for(i=0;i<PE_1y1.d0;++i) // incoming node is i
      for(uint k=0;k<PE_1y1.d2;++k) { // outgoing node is k
        bool obs = false;
        for(uint j=0;j<PE_1y1.d1;++j) { // outgoing obs. is j
          // Merge observations 
          if(PE_1y1(i,j,k) > 0.01) { 
            if(!obs) { 
              out << "\tN1_" << k << "-> N1_" << i << " [label = \""; 
              obs=true;
            }
            if (PE_1y1(i,j,k) < 0.99) 
            {out << "" << o[j] << ":" << std::setprecision(2) 
            << PE_1y1(i,j,k) << ",";}
          }
        }
        if(obs) out << "\",weight=50,arrowsize=0.3,labelfontsize=5.0,penwidth=0.5];\n";
      }
      // PE_01; E is true
      for(i=0;i<PE_01.d0;++i) // incoming node is i
        for(j=0;j<PE_01.d1;++j) { // outgoing node is j
          double value=PE_01(i,j);
          if(value > 0.01) {
            if(value > 0.99)
              out << "\tN1_" << j << ((Pe0(1,i)>0.99)? "-> E_":"-> N0_") << a0[i] << " [style=dashed,arrowsize=0.3,labelfontsize=5.0,penwidth=0.5];\n"; 
            else
              out << "\tN1_" << j << ((Pe0(1,i)>0.99)? "-> E_":"-> N0_") << a0[i] << " [label=\"" << std::setprecision(2) << value <<  "\",style=dashed,arrowsize=0.3,labelfontsize=5.0,penwidth=0.5];\n"; 
          }
        }
        
        out << "}\n";
        out.close();
        cout << "Done" << endl;
}


// Output the policy to dot format
// Currently only works for 1-level policies
void OutputDot(const char* filename, const char* problem, arr& Pa0, arr& P0y0) {
  std::vector<std::string> a, o;
  std::map<int, std::string> a0;
  OutputAoDot(Pa0, P0y0.d1, problem, a, o, a0);
  
  uint i,j;
  // Output graph
  ofstream out(filename);
  out << "digraph " << "ML_Controller_" << Pa0.d1 << "_" << 0 << " {\n"
    << "\trankdir=LR;\n"
    << "\tsize=\"8,5\"\n"
    << "\tnode [shape = circle];"; for(i=0;i<P0y0.d0;++i) {out << "N_" << a0[i] << " ";}
  out << ";\n"
    << "\tnode [shape = circle];\n";
  for(i=0;i<P0y0.d0;++i) // incoming node is i
    for(uint k=0;k<P0y0.d2;++k) { // outgoing node is k
      bool obs = false;
      for(j=0;j<P0y0.d1;++j) { // outgoing obs. is j
        // Merge observations 
        if(P0y0(i,j,k) > 0.01) { 
          if(!obs) { 
            out << "\tN_" << a0[k] << "-> N_" << a0[i] << "[ label = \""; 
            obs=true;
          }
          if (P0y0(i,j,k) > 0.99) out << "" << o[j] << ",";
          else {out << "" << o[j] << ":" << std::setprecision(2) << P0y0(i,j,k) << ",";}
        }
      }
      if(obs) out << "\"];\n";
    }
    
    out << "}\n";
    out.close();
}


// This function simply outputs the policy in an AMPL readable format.
// It's useful to debug the value of the policy.
void SetAMPLPolicy(arr& P0y0, arr& Pa0) { 
  uint i;
  
  // re-set the number of variables (to be independent of the .data file)
  cout << "set Nodes :=";
  for(i=0; i < P0y0.d0; i++) { cout << " " << i; } 
  cout << ";\n";
  
  arr P0ay0;
  P0ay0.resize(TUP(P0y0.d0, Pa0.d0, P0y0.d1, P0y0.d2));
  tensorEquation(P0ay0, P0y0, TUP(0,2,3), Pa0, TUP(1,3), 0);	
  tensorCheckCondNormalization(P0ay0,1);
  
  for(i=0; i < P0ay0.d0; i++) {
    for(uint j=0; j < P0ay0.d1; j++) {
      for(uint l=0; l < P0y0.d2; l++) { // switch the order of l and k since 
        for(uint k=0; k < P0ay0.d2; k++) { // the AMPL input is actually P0a0y
          cout << "\nlet x[" << i << ", " << j << ", " << l << ", " << k 
            << "] := "    << P0ay0(TUP(i,j,k,l))/*return4DNumber(P0ay0, i,j,k,l)*/ << ";";
        }
      }
    }
  }
  cout << endl;
}


