#include "mdp.h"
#include "mstep.h"
#include "infer.h"

#ifndef rescaleRewards
#  define rescaleRewards false
#endif

#undef useStructure
#define useStructure

double pomdpEM_lev1(
      const MDP& mdp,
      FSC_lev1& fsc,
      uint T,
      bool structuredEstep,
      bool maxMstep,
      bool adaptP0){
  
  CHECK(mdp.Px.nd==1 && mdp.Pxax.nd==3 &&
        mdp.Pyxa.nd==3 && mdp.Rax.nd==2 &&
        fsc.Pa0.nd==2 && fsc.P0y0.nd==3,"");
  
  uint
    dx=mdp.Pxax.d0,
    da=mdp.Pxax.d1,	
    dy=mdp.Pyxa.d0,
    d0=fsc.P0y0.d0;

  //----- rescale rewards if necessary
  arr mdp_Rax = mdp.Rax;
  double Rmin=mdp_Rax.min(),Rmax=mdp_Rax.max();
  if(rescaleRewards || (!maxMstep && Rmin<0.)){
    //if(!rescaleRewards) MT_MSG("can't handle neg rewards in case of exact M-step -- I'm enforcing rescaling of rewards!");
    for(uint i=0;i<mdp_Rax.N;i++) mdp_Rax.elem(i) = (mdp_Rax.elem(i)-Rmin)/(Rmax-Rmin);
  }else{
    Rmin=0.; Rmax=1.;
  }

  MT::timerStart();
  
  //----- define the factor model
  Variable x  (dx ,"state(t)");
  Variable y  (dy ,"observation(t)");
  Variable n0 (d0 ,"node0(t)");
  Variable a  (da ,"action(t)");
  Variable x_ (dx ,"state(t+1)");
  Variable y_ (dy ,"observation(t+1)");
  Variable n0_(d0 ,"node0(t+1)");
  //start
  Factor Fx  (TUP(x)        ,mdp.Px);
  Factor F0  (TUP(n0)       ,fsc.P0);
  //transition
  Factor Fa0 (TUP(a,n0)     ,fsc.Pa0);
  Factor Fxax(TUP(x_,a,x)   ,mdp.Pxax);
  Factor Fyxa(TUP(y_,x_,a)  ,mdp.Pyxa);
  Factor F0y0(TUP(n0_,y_,n0),fsc.P0y0);
  //reward
  Factor FRax(TUP(a,x)      ,mdp_Rax);  
  
  Factor Falpha(TUP(n0 ,x ));
  Factor Fbeta (TUP(n0_,x_));
  arr PT;
  double PR,ET;
  if(!structuredEstep){
    //----- collapse to unstructured model for generic inference
    FacA transitions=TUPLE(&Fa0, &Fxax, &Fyxa, &F0y0);
    //get transition matrix
    Factor F0x0x;
    eliminationAlgorithm(F0x0x,transitions,TUP(n0_,x_,n0,x));
    F0x0x.P.reshape(d0*dx,d0*dx);
    //get reward
    Factor FR0x;
    tensorProductMarginal(FR0x,Fa0,FRax,TUP(a));
    CHECK(FR0x.varIds==TUP(n0,x),"");
    FR0x.P.reshape(d0*dx);
    //get start
    Factor F0x;
    tensorProduct(F0x,F0,Fx);
    CHECK(F0x.varIds==TUP(n0,x),"");
    F0x.P.reshape(d0*dx);
    
    //----- E-STEP
    arr alpha,beta;
    inferMixLengthUnstructured(alpha,beta,PT,PR,ET,
                              F0x.P,FR0x.P,F0x0x.P,mdp.gamma,T);
    Falpha.set(alpha);
    Fbeta .set(beta);
  }else{
    FacA trans = TUPLE(&F0y0, &Fyxa, &Fxax, &Fa0);
    FacA newed;
    //eliminateVariable(trans,newed,a);
    //eliminateVariable(trans,newed,y_);
  
    FacA rewards = TUPLE(&FRax, &Fa0);
    //eliminateVariable(rewards,newed,a);
  
    inferMixLengthStructured(Falpha,Fbeta,PT,PR,ET,
                            TUP(n0 ,x ),TUP(n0_,x_),
                            TUPLE(&F0,&Fx),
                            rewards,
                            trans,
                            mdp.gamma,T);
    
    for(uint i=0;i<newed.N;i++) delete newed(i);
  }
  
  cout <<"E: " <<MT::timerRead(true) <<"sec, M: " <<std::flush;
  
  //----- M-STEP
  //consider the 2nd term (alpha*P_(x'|x)*beta)
  FacA twotimeslice = TUPLE(&Falpha, &Fa0, &Fxax, &Fyxa, &F0y0, &Fbeta);
  
  Factor X0y0_term2;
  eliminationAlgorithm(X0y0_term2,twotimeslice,TUP(n0_,y_,n0));
  Factor Xa0_term2;
  eliminationAlgorithm(Xa0_term2,twotimeslice,TUP(a,n0));
  
  //consider the 1st term (alpha*R_x)
  FacA immediateR   = TUPLE(&Falpha, &Fa0, &Fxax, &Fyxa, &F0y0, &FRax);

  Factor X0y0_term1;
  eliminationAlgorithm(X0y0_term1 ,immediateR,TUP(n0_,y_,n0));
  Factor Xa0_term1;
  eliminationAlgorithm(Xa0_term1,immediateR,TUP(a,n0));
    
  arr X0y0;
#if 0 //this parameter does not depend on immediate reward (only term2 is relevant)
  X0y0 = X0y0_term2.P*::exp(X0y0_term2.logP)
       + X0y0_term1.P*::exp(X0y0_term1.logP);
#else
  X0y0 = X0y0_term2.P*::exp(X0y0_term2.logP);
#endif
    
  arr Xa0;
  Xa0 = Xa0_term2.P*::exp(Xa0_term2.logP)
      + Xa0_term1.P*::exp(Xa0_term1.logP);
  
  //do the M-step!
  if(maxMstep){
    noisyMaxMstep(fsc.P0y0, X0y0, 1);
    noisyMaxMstep(fsc.Pa0 , Xa0 , 1);
    //if(adaptP0) noisyMaxMstep(fsc.P0   , X0.P   , 1);
  }else{
    standardMstep(fsc.P0y0, X0y0, 1);
    standardMstep(fsc.Pa0 , Xa0 , 1);
    //if(adaptP0) standardMstep(fsc.P0   , X0.P   , 1);
  }

#if 0 //comparison with old code
  //step 6: setup detailed (including y and a) factor graph over 2-time-slice
  //one big tensor
  arr A; A.resize(TUP(d0,dy,dx,da,d0,dx));
  uint _n0_=0, _y_=1, _x_=2, _a=3, _n0=4, _x=5;
  A = 1.;
  tensorMultiply(A,fsc.Pa0 ,TUP( _a ,_n0   ));
  tensorMultiply(A,mdp.Pxax,TUP( _x_,_a ,_x ));
  tensorMultiply(A,mdp.Pyxa,TUP( _y_,_x_,_a ));
  tensorMultiply(A,fsc.P0y0,TUP(_n0_,_y_,_n0));

  //multiply alpha
  tensorMultiply(A,alpha,TUP(_n0,_x));

  arr betaFac; betaFac.resize(TUP(d0,dx,da,dx)); betaFac=1.;
  tensorMultiply(betaFac,beta,TUP(0,1));
  arr rewardFac; rewardFac.resize(TUP(d0,dx,da,dx)); rewardFac=1.;
  tensorMultiply(rewardFac,mdp_Rax,TUP(2,3));
  rewardFac += betaFac;
  
  tensorMultiply(A,rewardFac,TUP(_n0_,_x_,_a,_x));

  // M-step: extract posterior marginals
  arr _X0y0(_n0_,_y_,_n0);
  arr _Xa0(_a ,_n0 );
  tensorMarginal(_Xa0 ,A,TUP(_a,_n0));
  tensorMarginal(_X0y0,A,TUP(_n0_,_y_,_n0));

  _X0y0.reshape(d0,dy,d0);
  _Xa0.reshape(da,d0);
  
  MT::save(X0y0,"z.1");
  MT::save(_X0y0,"z.2");
  
  cout <<"X0y0 error = " <<maxDiff(X0y0,_X0y0,0) <<endl;
  cout <<"Xa0 error = " <<maxDiff(Xa0,_Xa0,0) <<endl;
#endif

  tensorCheckCondNormalization(fsc.Pa0,1);
  tensorCheckCondNormalization(fsc.P0y0,1);
  
  //----- rest is cosmetics
  //report
  cout <<MT::timerRead() <<"sec," <<std::flush;
  cout <<" P(r=1)=" <<PR
       <<", Exp(T)=" <<ET <<"/"<<::log(PR)/::log(mdp.gamma)
       <<", Exp(R)=" <<(PR*(Rmax-Rmin)+Rmin)/(1.-mdp.gamma)
       <<endl;
 
  return (PR*(Rmax-Rmin)+Rmin);
}
