#ifndef MT_MSVC

#include <mdp/mdp.h>
#include "array.h"

void MatrixToArray(arr& X,Matrix M,uint d0,uint d1){
  CHECK(d0==(uint)M->num_rows,"given dimensionality not consistent");
  X.resize(d0,d1);
  X.setZero();
  uint i, j;
  for(i=0;i<(uint)M->num_rows;i++ ) {
    for(j=M->row_start[i]; j<(uint)(M->row_start[i]+M->row_length[i]); j++)
      X(i,M->col[j])= M->mat_val[j];
  }
}

int readTonyMDP(const char *filename, arr &Pxax, arr &Pyxa, arr &Px, arr &Rax,double &gamma){
  /*
    This routine returns 1 if the file is successfully parsed and 0 if not.
  */
  cout <<"parsing pomdp file `" <<filename <<"' using Cassandra's mdp library" <<endl;

  readMDP( (char*)filename );

  uint a,nx=gNumStates, ny=gNumObservations, na=gNumActions;
  cout <<"nx=" <<nx <<"  ny=" <<ny <<"  na=" <<na <<endl;

  cout <<"Paxy =\n"; for(a=0;a<na;a++){ cout <<"[a="<<a<<"]"; displayMatrix(R[a]); }
  cout <<"Paxx =\n"; for(a=0;a<na;a++){ cout <<"[a="<<a<<"]"; displayMatrix(P[a]); }
  cout <<"Q(a,x) =\n"; displayMatrix(Q);

  arr Paxy(na,nx,ny);
  for(a=0;a<na;a++) MatrixToArray(Paxy[a](),R[a],nx,ny);
  cout <<"Paxy =\n" <<Paxy <<endl;

  Pyxa.resize(ny,nx,na); //a permutation of Paxy
  Pyxa=1.;
  tensorMultiply(Pyxa,Paxy,TUP(2,1,0));
  cout <<"Pyax =\n" <<Pyxa <<endl;
  checkNormalization(Pyxa);

  arr Paxx(na,nx,nx);
  for(a=0;a<na;a++) MatrixToArray(Paxx[a](),P[a],nx,nx);
  cout <<"Paxx =\n" <<Paxx <<endl;
  Pxax.resize(nx,na,nx);
  Pxax=1.;
  tensorMultiply(Pxax,Paxx,TUP(1,2,0));
  cout <<"Pxax =\n" <<Pxax <<endl;
  checkNormalization(Pxax,1e-4);
  tensorCondNormalize(Pxax,1);

  Rax.resize(na,nx);
  MatrixToArray(Rax,Q,na,nx);
  cout <<"Rax =\n" <<Rax <<endl;

  Px.setCarray(gInitialBelief,nx);
  cout <<"Px =\n" <<Px <<endl;
  checkNormalization(Px,1e-4);
  tensorCondNormalize(Px,1);

  gamma = gDiscount;

  return 1;
}

#else

int readTonyMDP(const char *filename, arr &Pxax, arr &Pyxa, arr &Px, arr &Rax,double &gamma){
  HALT("reading Tony's POMDP file only works for Linux");
}

#endif

